#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__all__ = ["P115Path", "P115FileSystem"]

from asyncio import Lock as AsyncLock
from collections.abc import (
    AsyncIterable, AsyncIterator, Coroutine, Iterable, 
    Iterator, MutableMapping, 
)
from os import PathLike
from threading import Lock
from typing import overload, Any, ClassVar, Literal, Self

from dictattr import AttrDict
from dicttools import dict_key_to_lower_update
from errno2 import errno
from filewrap import Buffer, SupportsRead
from http_request import SupportsGeturl
from iterutils import run_gen_step
from p115pickcode import is_valid_pickcode
from yarl import URL

from ..client import check_response, P115Client
from ..exception import throw, P115BusyOSError
from ..type import P115URL
from ..tool import get_attr, get_ancestors, iterdir, normalize_attr_simple, P115QueryDB
from ..util import call_with_lock
from .fs_base import IDOrPathType, P115PathBase, P115FileSystemBase, AncestorDict


class P115Path(P115PathBase):
    __slots__ = ("fs", "attr")
    fs: P115FileSystem

    @overload
    def copy(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def copy(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def copy(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        return self.fs.copy(
            self, 
            to_dir=to_dir, 
            refresh=refresh, 
            async_=async_, 
            **request_kwargs, 
        )

    @overload
    def mkdir(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Self:
        ...
    @overload
    def mkdir(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, Self]:
        ...
    def mkdir(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Self | Coroutine[Any, Any, Self]:
        def gen_step():
            if not self.is_dir():
                throw(errno.ENOTDIR, self.attr)
            attr = yield self.fs.mkdir(
                self, 
                name=name, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            return type(self)(self.fs, attr)
        return run_gen_step(gen_step, async_)

    @overload
    def move(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Self:
        ...
    @overload
    def move(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, Self]:
        ...
    def move(
        self, 
        /, 
        to_dir: IDOrPathType, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Self | Coroutine[Any, Any, Self]:
        def gen_step():
            yield self.fs.move(
                self, 
                to_dir=to_dir, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            return self
        return run_gen_step(gen_step, async_)

    @overload
    def remove(
        self, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Self:
        ...
    @overload
    def remove(
        self, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, Self]:
        ...
    def remove(
        self, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Self | Coroutine[Any, Any, Self]:
        def gen_step():
            yield self.fs.remove(
                self, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            return self
        return run_gen_step(gen_step, async_)

    @overload
    def rename(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Self:
        ...
    @overload
    def rename(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, Self]:
        ...
    def rename(
        self, 
        /, 
        name: str, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Self | Coroutine[Any, Any, Self]:
        def gen_step():
            yield self.fs.rename(
                self, 
                name, 
                refresh=refresh, 
                async_=async_, # type: ignore
                **request_kwargs, 
            )
            return self
        return run_gen_step(gen_step, async_)


class P115FileSystem(P115FileSystemBase[P115Path]):
    STAT_MODE: ClassVar = 0o777
    NO_INCREMENT: ClassVar = False
    path_class: ClassVar = P115Path
    """对 115 网盘模拟文件系统的操作

    :param client: 115 客户端或 cookies
    :param refresh: 是否总是刷新，如果是，则不使用缓存
    :param id_to_readdir: 缓存用的字典，映射目录 id 到在此目录中的文件或目录的另一个字典，后者是文件或目录的 id 到它的信息字典的字典
    """
    def __init__(
        self, 
        /, 
        client: str | PathLike | P115Client, 
        refresh: bool = False, 
        id_to_readdir: None | int | dict[int, dict[int, MutableMapping]] = None, 
    ):
        super().__init__(client, refresh=refresh, id_to_readdir=id_to_readdir)
        self._fs_lock = Lock()
        self._fs_alock = AsyncLock()

    @overload
    def _get_attr_by_id(
        self, 
        id: int, 
        /, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def _get_attr_by_id(
        self, 
        id: int, 
        /, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def _get_attr_by_id(
        self, 
        id: int, 
        /, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        if refresh is None:
            refresh = self.refresh
        def gen_step():
            if id == 0:
                return self._get_root_attr()
            id_to_attr = self.id_to_attr
            if not refresh:
                if attr := id_to_attr.get(id):
                    return attr
            attr = yield get_attr(
                self.client, 
                id, 
                async_=async_, 
                **request_kwargs, 
            )
            if attr_old := id_to_attr.get(id):
                attr_old.update(attr)
                if id_to_readdir := self.id_to_readdir:
                    if attr_old["parent_id"] != attr["parent_id"]:
                        try:
                            id_to_readdir[attr_old["parent_id"]].pop(id, None)
                        except KeyError:
                            pass
                        try:
                            id_to_readdir[attr["parent_id"]][id] = attr_old
                        except KeyError:
                            pass
                attr = attr_old
            else:
                id_to_attr[id] = attr
            return attr
        return run_gen_step(gen_step, async_)

    @overload
    def _get_ancestors_by_cid(
        self, 
        cid: int = 0, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> list[AncestorDict]:
        ...
    @overload
    def _get_ancestors_by_cid(
        self, 
        cid: int = 0, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, list[AncestorDict]]:
        ...
    def _get_ancestors_by_cid(
        self, 
        cid: int = 0, 
        /, 
        refresh: None | bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> list[AncestorDict] | Coroutine[Any, Any, list[AncestorDict]]:
        """获取某个目录 id 对应的祖先节点信息（包括自身）
        """
        def gen_step():
            if cid == 0:
                return [{"id": 0, "parent_id": 0, "name": ""}]
            if not refresh:
                try:
                    return P115QueryDB(self.con).get_ancestors(cid)
                except (ValueError, FileNotFoundError):
                    pass
            return (yield get_ancestors(
                self.client, 
                cid, 
                id_to_dirnode=self.id_to_dirnode, 
                ensure_file=False, 
                refresh=True, 
                async_=async_, 
                **request_kwargs, 
            ))
        return run_gen_step(gen_step, async_)

    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Iterator[dict]:
        ...
    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> AsyncIterator[dict]:
        ...
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Iterator[dict] | AsyncIterator[dict]:
        return iterdir(
            self.client, 
            id, 
            normalize_attr=normalize_attr_simple, 
            id_to_dirnode=self.id_to_dirnode, 
            async_=async_, 
            **request_kwargs, 
        )

    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> P115URL:
        ...
    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, P115URL]:
        ...
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> P115URL | Coroutine[Any, Any, P115URL]:
        "获取下载链接"
        if refresh is None:
            refresh = self.refresh
        def gen_step():
            headers = request_kwargs["headers"] = dict(request_kwargs.get("headers") or ())
            dict_key_to_lower_update(headers)
            user_agent = headers.pop("user-agent", "")
            app = "android"
            if isinstance(id_or_path, int):
                pickcode = self.client.to_pickcode(id_or_path)
            elif isinstance(id_or_path, str) and is_valid_pickcode(id_or_path):
                pickcode = id_or_path
            else:
                attr = yield self.get_attr(
                    id_or_path, 
                    pid=pid, 
                    refresh=refresh, 
                    async_=async_, 
                    **request_kwargs, 
                )
                if attr["is_dir"]:
                    throw(errno.EISDIR, attr)
                pickcode = attr["pickcode"]
                if attr.get("is_collect", False):
                    if attr["size"] <= 1024 * 1024 * 200:
                        app = "web"
                    else:
                        raise throw(errno.EIO, f"file {id_or_path!r} has been censored")
            return self.client.download_url(
                pickcode, 
                user_agent=user_agent, 
                app=app, 
                async_=async_, 
                **request_kwargs, 
            )
        return run_gen_step(gen_step, async_)

    @overload
    def copy(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def copy(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def copy(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        """复制文件或目录
        """
        def gen_step():
            id = yield self.get_id(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            to_cid = yield self.get_id(
                to_dir, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            lock = self._fs_alock if async_ else self._fs_lock
            while True:
                resp = yield call_with_lock(
                    lock, 
                    self.client.fs_copy_app, 
                    id, 
                    pid=to_cid, 
                    async_=async_, 
                    **request_kwargs, 
                )
                try:
                    check_response(resp)
                    self.id_to_readdir.pop(to_cid, None)
                    return resp
                except P115BusyOSError:
                    continue
        return run_gen_step(gen_step, async_)

    @overload
    def mkdir(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def mkdir(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def mkdir(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        "创建目录"
        def gen_step():
            cid = yield self.get_id(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            resp = yield self.client.fs_mkdir_app(
                name, 
                pid=cid, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            info = resp["data"]
            fid = int(info["category_id"])
            cname = info["category_name"]
            attr: dict = AttrDict(
                id=fid, 
                parent_id=cid, 
                is_dir=True, 
                name=cname, 
                pickcode=info["pick_code"], 
                ctime=int(info["ptime"]), 
                mtime=int(info["utime"]), 
            )
            children = self.id_to_readdir.get(cid)
            if children is not None:
                children[fid] = self.id_to_attr[fid] = attr
                self._pid_name_to_attr[(cid, attr["name"])] = attr
                self.id_to_dirnode[fid] = (cname, cid)
            return attr
        return run_gen_step(gen_step, async_)

    @overload
    def move(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def move(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def move(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        to_dir: IDOrPathType = "", 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        "移动文件或目录"
        def gen_step():
            attr = yield self.get_attr(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            fid = attr["id"]
            to_cid = yield self.get_id(
                to_dir, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            lock = self._fs_alock if async_ else self._fs_lock
            while True:
                resp = yield call_with_lock(
                    lock, 
                    self.client.fs_move_app, 
                    fid, 
                    pid=to_cid, 
                    async_=async_, 
                    **request_kwargs, 
                )
                try:
                    check_response(resp)
                    id_to_readdir = self.id_to_readdir
                    try:
                        id_to_readdir[attr["parent_id"]].pop(fid)
                    except KeyError:
                        pass
                    attr["parent_id"] = to_cid
                    try:
                        id_to_readdir[to_cid][fid] = attr
                    except KeyError:
                        pass
                    if attr["is_dir"]:
                        self.id_to_dirnode[fid] = (attr["name"], to_cid)
                    return attr
                except P115BusyOSError:
                    continue
        return run_gen_step(gen_step, async_)

    @overload
    def remove(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def remove(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def remove(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        "删除文件或目录"
        def gen_step():
            attr = yield self.get_attr(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            fid = attr["id"]
            lock = self._fs_alock if async_ else self._fs_lock
            while True:
                resp = yield call_with_lock(
                    lock, 
                    self.client.fs_delete_app, 
                    fid, 
                    async_=async_, 
                    **request_kwargs, 
                )
                try:
                    check_response(resp)
                    self.id_to_dirnode.pop(fid, None)
                    self.id_to_attr.pop(fid, None)
                    self.id_to_readdir.pop(fid, None)
                    try:
                        cid = attr["parent_id"]
                        self.id_to_readdir[cid].pop(fid, None)
                    except KeyError:
                        pass
                    return attr
                except P115BusyOSError:
                    continue
        return run_gen_step(gen_step, async_)

    @overload
    def rename(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def rename(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def rename(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        name: str, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        """重命名文件或路径
        """
        assert name
        def gen_step():
            attr = yield self.get_attr(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            fid = attr["id"]
            resp = yield self.client.fs_rename_app(
                (fid, name), 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            if data := resp["data"]:
                attr["name"] = data[str(fid)]
            return attr
        return run_gen_step(gen_step, async_)

    @overload
    def upload(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        file: ( Buffer | str | PathLike | URL | SupportsGeturl | 
                SupportsRead | Iterable[Buffer] ) = b"", 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> AttrDict:
        ...
    @overload
    def upload(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        file: ( Buffer | str | PathLike | URL | SupportsGeturl | 
                SupportsRead | Iterable[Buffer] | AsyncIterable[Buffer] ) = b"", 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, AttrDict]:
        ...
    def upload(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        file: ( Buffer | str | PathLike | URL | SupportsGeturl | 
                SupportsRead | Iterable[Buffer] | AsyncIterable[Buffer] ) = b"", 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> AttrDict | Coroutine[Any, Any, AttrDict]:
        "上传文件到目录"
        def gen_step():
            attr = yield self.get_attr(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            if not attr["is_dir"]:
                throw(errno.ENOTDIR, attr)
            cid = attr["id"]
            if isinstance(file, (Buffer, str, PathLike, URL, SupportsGeturl, SupportsRead)):
                resp = yield self.client.upload_file(
                    file, 
                    pid=cid, 
                    filename=filename, 
                    filesha1=filesha1, 
                    filesize=filesize, 
                    partsize=10485760, 
                    async_=async_, 
                    **request_kwargs, 
                )
                check_response(resp)
                if resp.get("request") == "upload":
                    info = resp["data"]
                    attr = AttrDict(
                        id=int(info["id"]), 
                        parent_id=cid, 
                        name=info["filename"], 
                        is_dir=False, 
                        sha1=info["filesha1"], 
                        size=int(info["filesize"]), 
                        pickcode=info["pickcode"], 
                    )
                else:
                    info = resp["data"]
                    attr = AttrDict(
                        id=int(info["file_id"]), 
                        parent_id=cid, 
                        name=info["file_name"], 
                        is_dir=False, 
                        sha1=info["sha1"], 
                        size=int(info["file_size"]), 
                        pickcode=info["pick_code"], 
                    )
            else:
                resp = yield self.client.upload_file_sample(
                    file, # type: ignore
                    filename=filename, 
                    pid=cid, 
                    async_=async_, # type: ignore
                    **request_kwargs, 
                )
                check_response(resp)
                info = resp["data"]
                ctime = int(info["file_ptime"])
                attr = AttrDict(
                    id=int(info["file_id"]), 
                    parent_id=cid, 
                    name=info["file_name"], 
                    is_dir=False, 
                    sha1=info["sha1"], 
                    size=int(info["file_size"]), 
                    pickcode=info["pick_code"], 
                    ctime=ctime, 
                    mtime=ctime, 
                )
            children = self.id_to_readdir.get(cid)
            if children is not None:
                fid = attr["id"]
                children[fid] = self.id_to_attr[fid] = attr
                self._pid_name_to_attr[(cid, attr["name"])] = attr
            return attr
        return run_gen_step(gen_step, async_)

# TODO: 实现 search 方法，以及可以设定 desc、label 等
# TODO: 允许手动指定一个 escape 方法
# TODO: 增加多种方法: copy_many, move_many, rename_many, remove_many
# TODO: 增加方法 copyfile、renamefile，可以改变用不同名字
# TODO: 增加方法，moves，可以改名和移动，且文件的后缀名可以不同
# TODO: 增加方法，copys，如果是目录，复制不可改名，但是文件则可以，而且后缀可以不同



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["deauth_open"]
__doc__ = "这个模块提供了一些和账号状况有关的函数"

from collections.abc import Callable, Coroutine
from os import PathLike
from typing import overload, Any, Literal

from iterutils import run_gen_step

from ..client import check_response, P115Client


@overload
def deauth_open(
    client: str | PathLike | P115Client, 
    predicate: None | Callable = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> None:
    ...
@overload
def deauth_open(
    client: str | PathLike | P115Client, 
    predicate: None | Callable = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, None]:
    ...
def deauth_open(
    client: str | PathLike | P115Client, 
    predicate: None | Callable = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> None | Coroutine[Any, Any, None]:
    """批量解绑开放应用

    :param client: 115 客户端或 cookies
    :param predicate: 筛选条件
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    def gen_step():
        resp = yield client.login_open_auth_list(
            async_=async_, 
            **request_kwargs, 
        )
        check_response(resp)
        for info in filter(predicate, resp["data"]):
            yield client.login_open_deauth(
                info["auth_id"], 
                async_=async_, 
                **request_kwargs, 
            )
    return run_gen_step(gen_step, async_)




#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__all__ = ["P115SharePath", "P115ShareFileSystem"]

from collections.abc import (
    AsyncIterator, Coroutine, Iterable, Iterator, MutableMapping, 
)
from datetime import datetime
from functools import cached_property
from os import PathLike
from typing import overload, Any, ClassVar, Literal

from iterutils import run_gen_step

from ..client import check_response, P115Client
from ..type import P115URL
from ..tool import share_iterdir, normalize_attr_simple
from ..util import share_extract_payload
from .fs_base import IDOrPathType, P115PathBase, P115FileSystemBase


class P115SharePath(P115PathBase):
    __slots__ = ("fs", "attr")
    fs: P115ShareFileSystem

    @overload
    def receive(
        self, 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def receive(
        self, 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def receive(
        self, 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        return self.fs.receive(
            {
                "share_code": self.share_code, 
                "receive_code": self.receive_code, 
                "file_id": self.id, 
                "cid": to_pid, 
            }, 
            async_=async_, 
            **request_kwargs, 
        )


class P115ShareFileSystem(P115FileSystemBase):
    path_class: ClassVar = P115SharePath

    def __init__(
        self, 
        /, 
        client: str | PathLike | P115Client, 
        share_code: str, 
        receive_code: None | str = None, 
        refresh: bool = False, 
        id_to_readdir: None | int | dict[int, dict[int, MutableMapping]] = None, 
    ):
        super().__init__(client, refresh=refresh, id_to_readdir=id_to_readdir)
        self.share_code = share_code
        self.receive_code = receive_code or ""

    @classmethod
    def from_url(
        cls, 
        /, 
        client: str | PathLike | P115Client, 
        url: str, 
        refresh: bool = False, 
        id_to_readdir: None | int | dict[int, dict[int, MutableMapping]] = None, 
    ) -> P115ShareFileSystem:
        return cls(
            client, 
            **share_extract_payload(url), 
            refresh=refresh, 
            id_to_readdir=id_to_readdir, 
        )

    @cached_property
    def create_time(self, /) -> datetime:
        "分享的创建时间"
        return datetime.fromtimestamp(self.create_timestamp)

    @cached_property
    def create_timestamp(self, /) -> int:
        "分享的创建时间"
        return int(self.share_info["create_time"])

    @cached_property
    def snap_id(self, /) -> int:
        "获取这个分享的 id"
        return int(self.share_info["snap_id"])

    @cached_property
    def share_user_id(self, /) -> int:
        "获取分享者的用户 id"
        return int(self.share_data["userinfo"]["user_id"])

    @property
    def share_data(self, /) -> dict:
        "获取分享的首页数据"
        return self.get_share_data()

    @property
    def share_info(self, /) -> dict:
        "获取分享信息"
        return self.share_data["share_info"]

    @overload
    def get_share_data(
        self, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def get_share_data(
        self, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def get_share_data(
        self, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        def gen_step():
            resp = yield self.client.share_snap(
                {
                    "share_code": self.share_code, 
                    "receive_code": self.receive_code, 
                    "limit": 1, 
                }, 
                async_=async_, 
                **request_kwargs, 
            )
            return resp["data"]
        return run_gen_step(gen_step, async_)

    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> P115URL:
        ...
    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, P115URL]:
        ...
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> P115URL | Coroutine[Any, Any, P115URL]:
        "获取下载链接"
        def gen_step():
            id = yield self.get_id(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
            return self.client.share_download_url(
                {
                    "share_code": self.share_code, 
                    "receive_code": self.receive_code, 
                    "file_id": id, 
                }, 
                async_=async_, 
                **request_kwargs, 
            )
        return run_gen_step(gen_step, async_)

    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Iterator[MutableMapping]:
        ...
    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> AsyncIterator[MutableMapping]:
        ...
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Iterator[MutableMapping] | AsyncIterator[MutableMapping]:
        return share_iterdir(
            self.client, 
            share_code=self.share_code, 
            receive_code=self.receive_code, 
            cid=id, 
            normalize_attr=normalize_attr_simple, 
            id_to_dirnode=self.id_to_dirnode, 
            async_=async_, 
            **request_kwargs, 
        )

    @overload
    def receive(
        self, 
        ids: int | str | Iterable[int | str], 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def receive(
        self, 
        ids: int | str | Iterable[int | str], 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def receive(
        self, 
        ids: int | str | Iterable[int | str], 
        /, 
        to_pid: int = 0, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        """接收分享文件到网盘

        :param ids: 要转存到文件 id（这些 id 归属分享链接）
        :param to_pid: 你的网盘的一个目录 id（这个 id 归属你的网盘）
        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 接口响应
        """
        return check_response(self.client.share_receive(
            {
                "share_code": self.share_code, 
                "receive_code": self.receive_code, 
                "file_id": ids if isinstance(ids, (int, str)) else ",".join(map(str, ids)), 
                "cid": to_pid, 
            }, 
            async_=async_, 
            **request_kwargs, 
        ))

# TODO: 如果是自己的分享，可以尝试获取分享码（增加一个方法，用来拉取分享码，增加一个方法，可以从自己的分享中查询某个分享的信息）
# TODO: 增加 search 方法
# TODO: share_download_url 增加 user_agent 参数（网页版接口可能需要？？？）



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "updatedb_initdb", "updatedb", "updatedb_life_iter", 
    "updatedb_history_iter", "P115QueryDB", 
]
__doc__ = "这个模块提供了一些和更新数据库有关的函数"

from collections.abc import (
    AsyncIterator, Callable, Collection, Coroutine, Iterable, Iterator, Sequence, 
)
from math import inf
from os import PathLike
from sqlite3 import register_adapter, register_converter, Connection, Cursor
from time import time
from typing import cast, overload, Any, Literal
from warnings import warn

from asynctools import ensure_async
from errno2 import errno
from iter_collect import grouped_mapping
from iterutils import (
    bfs_gen, chunked, foreach, run_gen_step, run_gen_step_iter, 
    with_iter_next, Yield, 
)
from orjson import dumps, loads
from p115pickcode import to_id
from posixpatht import path_is_dir_form, escape, splits
from sqlitetools import connect, execute, find, query, upsert_items

from ..client import P115Client, P115Warning
from ..util import posix_escape_name
from .attr import get_ancestors, normalize_attr_simple
from .history import iter_history_list
from .iterdir import iterdir, iter_nodes_using_event, traverse_tree
from .life import iter_life_behavior_list


register_adapter(list, dumps)
register_adapter(dict, dumps)
register_converter("JSON", loads)


def updatedb_initdb(con: Connection | Cursor, /) -> Cursor:
    """初始化数据库，然后返回游标
    """
    sql = """\
-- 修改日志模式为 WAL (Write Ahead Log)
PRAGMA journal_mode = WAL;

-- data 表，用来保存数据
CREATE TABLE IF NOT EXISTS data (
    id INTEGER NOT NULL PRIMARY KEY,      -- 主键
    parent_id INTEGER NOT NULL DEFAULT 0, -- 上级目录的 id
    name TEXT NOT NULL,                   -- 名字
    sha1 TEXT NOT NULL DEFAULT '',        -- 文件的 sha1 哈希值
    size INTEGER NOT NULL DEFAULT 0,      -- 文件大小
    pickcode TEXT NOT NULL DEFAULT '',    -- 提取码，下载等操作时需要用到
    is_dir INTEGER NOT NULL DEFAULT 1 CHECK(is_dir IN (0, 1)), -- 是否目录
    is_alive INTEGER NOT NULL DEFAULT 1 CHECK(is_alive IN (0, 1)), -- 是否存活（存活即是不是删除状态）
    extra BLOB DEFAULT NULL,              -- 额外的数据
    created_at TIMESTAMP DEFAULT (unixepoch('subsec')), -- 创建时间
    updated_at TIMESTAMP DEFAULT (CAST(STRFTIME('%s', 'now') AS INTEGER))  -- 更新时间
);

-- life 表，用来保存操作事件
CREATE TABLE IF NOT EXISTS life (
    id INTEGER NOT NULL PRIMARY KEY, -- 文件或目录的 id
    data JSON NOT NULL, -- 数据
    created_at TIMESTAMP DEFAULT (unixepoch('subsec')) -- 创建时间
);

-- history 表，用来保存历史记录
CREATE TABLE IF NOT EXISTS history (
    id INTEGER NOT NULL PRIMARY KEY, -- 文件或目录的 id
    data JSON NOT NULL, -- 数据
    created_at TIMESTAMP DEFAULT (unixepoch('subsec')) -- 创建时间
);

-- 索引
CREATE INDEX IF NOT EXISTS idx_data_pid ON data(parent_id);
CREATE INDEX IF NOT EXISTS idx_data_utime ON data(updated_at);

-- data 表的记录发生更新，自动更新它的更新时间
CREATE TRIGGER IF NOT EXISTS trg_data_update
AFTER UPDATE ON data
FOR EACH ROW
BEGIN
    SELECT CASE
        WHEN NEW.updated_at < OLD.updated_at THEN RAISE(IGNORE)
    END;
    UPDATE data SET updated_at = CAST(STRFTIME('%s', 'now') AS INTEGER) WHERE id = NEW.id AND NEW.updated_at = OLD.updated_at;
END;

-- fs_event 表，用来保存文件系统变更（由 data 表触发）
CREATE TABLE IF NOT EXISTS fs_event (
    id INTEGER PRIMARY KEY AUTOINCREMENT, -- 事件 id
    event TEXT NOT NULL,                  -- 事件类型：add（增）、remove（删）、rename（改名）、move（移动）
    file_id INTEGER NOT NULL,             -- 文件或目录的 id，此 id 必在 `data` 表中
    pid0 INTEGER NOT NULL DEFAULT -1,     -- 变更前上级目录的 id
    pid1 INTEGER NOT NULL DEFAULT -1,     -- 变更后上级目录的 id
    name0 TEXT NOT NULL DEFAULT '',       -- 变更前的名字
    name1 TEXT NOT NULL DEFAULT '',       -- 变更后的名字
    created_at TIMESTAMP DEFAULT (unixepoch('subsec')) -- 创建时间
);

-- data 表发生插入
CREATE TRIGGER IF NOT EXISTS trg_data_insert
AFTER INSERT ON data
FOR EACH ROW
BEGIN
    INSERT INTO fs_event(event, file_id, pid1, name1) VALUES (
        'add', NEW.id, NEW.parent_id, NEW.name
    );
END;

-- data 表发生还原
CREATE TRIGGER IF NOT EXISTS trg_data_revoke
AFTER UPDATE ON data
FOR EACH ROW WHEN (NOT OLD.is_alive AND NEW.is_alive)
BEGIN
    INSERT INTO fs_event(event, file_id, pid1, name1) VALUES (
        'add', NEW.id, NEW.parent_id, NEW.name
    );
END;

-- data 表发生移除
CREATE TRIGGER IF NOT EXISTS trg_data_remove
AFTER UPDATE ON data
FOR EACH ROW WHEN (OLD.is_alive AND NOT NEW.is_alive)
BEGIN
    INSERT INTO fs_event(event, file_id, pid0, name0) VALUES (
        'remove', OLD.id, OLD.parent_id, OLD.name
    );
END;

-- data 表发生改名或移动
CREATE TRIGGER IF NOT EXISTS trg_data_change
AFTER UPDATE ON data
FOR EACH ROW WHEN (OLD.is_alive AND NEW.is_alive)
BEGIN
    INSERT INTO fs_event(event, file_id, pid0, pid1, name0, name1)
    SELECT
        'move', OLD.id, OLD.parent_id, NEW.parent_id, OLD.name, OLD.name
    WHERE OLD.parent_id != NEW.parent_id;

    INSERT INTO fs_event(event, file_id, pid0, pid1, name0, name1) 
    SELECT * FROM (
        SELECT
            'rename', NEW.id, NEW.parent_id, NEW.parent_id, OLD.name, NEW.name
        WHERE OLD.name != NEW.name
    );
END;"""
    return con.executescript(sql)


def wrap_async(func, async_: bool = False, /, threaded: bool = False):
    if async_:
        return ensure_async(func, threaded=threaded)
    else:
        return func


def _init_client(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    init_sql: None | str = None, 
) -> tuple[P115Client, Connection | Cursor]:
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if client.login_app() in ("web", "desktop", "harmony"):
        warn(
            f'app within ("web", "desktop", "harmony") is not recommended, it will be replaced by "apple_tv" cookies', 
            category=P115Warning, 
        )
        client.login_another_app("apple_tv", replace=True)
    if not dbfile:
        dbfile = f"p115db-{client.user_id}.db"
    if isinstance(dbfile, (Connection, Cursor)):
        con = dbfile
    else:
        con = connect(dbfile, check_same_thread=False, timeout=inf)
        if init_sql is None:
            updatedb_initdb(con)
        elif init_sql:
            con.executescript(init_sql)
    return client, con


def has_id(con: Connection | Cursor, id: int, /) -> int:
    sql = "SELECT 1 FROM data WHERE id = ? AND is_alive"
    return find(con, sql, (id,), default=0)


def event_normalize_attr(event: dict, /) -> dict:
    sha1 = event["sha1"]
    return {
        "id": int(event["file_id"]), 
        "parent_id": int(event["parent_id"]), 
        "name": event["file_name"], 
        "sha1": sha1, 
        "size": int(event.get("file_size") or 0), 
        "pickcode": event["pick_code"], 
        "is_dir": not sha1, 
        "is_alive": event["type"] != 22, 
        "updated_at": int(event["create_time"]), 
    }


@overload
def load_missing_ancestors(
    client: P115Client, 
    con: Connection | Cursor, 
    attrs: list[dict], 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> list[dict]:
    ...
@overload
def load_missing_ancestors(
    client: P115Client, 
    con: Connection | Cursor, 
    attrs: list[dict], 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, list[dict]]:
    ...
def load_missing_ancestors(
    client: P115Client, 
    con: Connection | Cursor, 
    attrs: list[dict], 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> list[dict] | Coroutine[Any, Any, list[dict]]:
    def gen_step():
        seen_ids: set[int] = {a["id"] for a in attrs}
        ancestors: list[dict] = []
        add_to_seen = seen_ids.add
        add_ancestor = ancestors.append
        def add(attr: dict, /):
            add_to_seen(attr["id"])
            add_ancestor(attr)
        while pids := [
            pid for a in attrs 
            if (pid := a["parent_id"]) and not (pid in seen_ids or has_id(con, pid))
        ]:
            yield foreach(
                add, 
                iter_nodes_using_event(
                    client, 
                    pids, 
                    type="doc", 
                    normalize_attr=event_normalize_attr, 
                    id_to_dirnode=..., 
                    cooldown=cooldown, 
                    app=app, 
                    async_=async_, 
                    **request_kwargs, 
                ), 
            )
        return ancestors
    return run_gen_step(gen_step, async_)


@overload
def updatedb(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    cid: int | str = 0, 
    recursive: bool = True, 
    max_workers: None | int = None, 
    max_files: int | None = 0, 
    max_dirs: int | None = 0, 
    app: str = "android", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> int:
    ...
@overload
def updatedb(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    cid: int | str = 0, 
    recursive: bool = True, 
    max_workers: None | int = None, 
    max_files: int | None = 0, 
    max_dirs: int | None = 0, 
    app: str = "android", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, int]:
    ...
def updatedb(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    cid: int | str = 0, 
    recursive: bool = True, 
    max_workers: None | int = None, 
    max_files: int | None = 0, 
    max_dirs: int | None = 0, 
    app: str = "android", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> int | Coroutine[Any, Any, int]:
    """对某个目录执行一次拉取，以更新 SQLite 数据

    :param client: 115 网盘客户端对象
    :param dbfile: 数据库文件路径，如果为 None，则自动确定
    :param cid: 目录的 id 或 pickcode
    :param recursive: 如果为 True，则拉取所有以之为祖先（先驱）节点的节点信息；否则，拉取所有以之为父（前驱）节点的节点信息
    :param max_workers: 最大并发数，如果为 None 或 < 0 则自动确定，如果为 0 则单工作者惰性执行
    :param max_files: 估计最大存在的文件数，<= 0 时则无限
    :param max_dirs: 估计最大存在的目录数，<= 0 时则无限
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 返回总共影响到数据行数，即所有 DML SQL 执行后，游标的 ``.rowcount`` 累加
    """
    client, con = _init_client(client, dbfile)
    upsert = wrap_async(upsert_items, async_, threaded=True)
    cid = to_id(cid)
    def gen_step():
        total = 0
        if recursive:
            start_t = int(time())
            try:
                if cid and not has_id(con, cid):
                    ancestors = yield get_ancestors(
                        client, 
                        cid, 
                        id_to_dirnode=..., 
                        ensure_file=False, 
                        app=app, 
                        async_=async_, 
                        **request_kwargs, 
                    )
                    if ancestors:
                        if ancestors[0]["id"] == 0:
                            ancestors = ancestors[1:]
                        if ancestors:
                            to_pickcode = client.to_pickcode
                            for a in ancestors:
                                a["pickcode"] = to_pickcode(a["id"], "fa")
                            total += (yield upsert(
                                con, 
                                ancestors, 
                                {"is_alive": 1}, 
                                commit=True, 
                            )).rowcount
                with with_iter_next(chunked(
                    traverse_tree(
                        client, 
                        cid, 
                        id_to_dirnode=..., 
                        max_workers=max_workers, 
                        max_files=max_files, 
                        max_dirs=max_dirs, 
                        app=app, 
                        async_=async_, 
                        **request_kwargs, 
                    ), 
                    1000, 
                )) as get_next:
                    while True:
                        batch = yield get_next()
                        total += (yield upsert(con, batch, {"is_alive": 1}, commit=True)).rowcount
                if cid:
                    clean_sql = f"""\
UPDATE data SET is_alive = 0 WHERE id in (
    WITH ids(id) AS (
        SELECT id FROM data WHERE parent_id = {cid} AND is_alive AND updated_at < :start_t
        UNION ALL
        SELECT data.id FROM ids JOIN data ON (ids.id = data.parent_id) WHERE is_alive AND updated_at < :start_t
    )
    SELECT id FROM ids
);"""
                else:
                    clean_sql = "UPDATE data SET is_alive = 0 WHERE is_alive AND updated_at < :start_t"
                total += (yield wrap_async(execute, async_, threaded=True)(
                    con, 
                    clean_sql, 
                    {"start_t": start_t}, 
                    commit=True, 
                )).rowcount
            except FileNotFoundError:
                if cid:
                    clean_sql = f"""\
UPDATE data SET is_alive = 0 WHERE id in (
    WITH ids(id) AS (
        SELECT id FROM data WHERE parent_id = {cid} AND is_alive
        UNION ALL
        SELECT data.id FROM ids JOIN data ON (ids.id = data.parent_id) WHERE is_alive
    )
    SELECT id FROM ids
);"""
                else:
                    clean_sql = "UPDATE data SET is_alive = 0 WHERE is_alive"
                total = (yield wrap_async(execute, async_, threaded=True)(
                    con, 
                    clean_sql, 
                    commit=True, 
                )).rowcount
        else:
            id_to_dirnode: dict[int, tuple[str, int]] = {}
            seen_ids: set[int] = set()
            try:
                with with_iter_next(chunked(iterdir(
                    client, 
                    cid, 
                    normalize_attr=normalize_attr_simple, 
                    id_to_dirnode=id_to_dirnode, 
                    raise_for_changed_count=True, 
                    app=app, 
                    cooldown=0.5, 
                    max_workers=max_workers, 
                    async_=async_, 
                    **request_kwargs, 
                ), 1000)) as get_next:
                    while True:
                        batch = yield get_next()
                        total += (yield upsert(
                            con, 
                            batch, 
                            extras={"is_alive": 1}, 
                            fields=("id", "parent_id", "name", "sha1", "size", "pickcode", "is_dir"), 
                            commit=True, 
                        )).rowcount
                        seen_ids.update(a["id"] for a in batch)
                if id_to_dirnode:
                    to_pickcode = client.to_pickcode
                    total += (yield upsert(
                        con, 
                        [
                            {"id": id, "name": name, "parent_id": pid, "ancestors": to_pickcode(id, "fa")} 
                            for id, (name, pid) in id_to_dirnode.items()
                            if id not in seen_ids
                        ], 
                        {"is_alive": 1}, 
                        commit=True, 
                    )).rowcount
                clean_sql = "UPDATE data SET is_alive = 0 WHERE is_alive and parent_id = ?"
                if seen_ids:
                    clean_sql += " AND id NOT IN (%s)" % ",".join(map(str, seen_ids))
                total += (yield wrap_async(execute, async_, threaded=True)(
                    con, 
                    clean_sql, 
                    (cid,), 
                    commit=True, 
                )).rowcount
            except FileNotFoundError:
                clean_sql = "UPDATE data SET is_alive = 0 WHERE is_alive and parent_id = ?"
                total = (yield wrap_async(execute, async_, threaded=True)(
                    con, 
                    clean_sql, 
                    (cid,), 
                    commit=True, 
                )).rowcount
        return total
    return run_gen_step(gen_step, async_)


@overload
def updatedb_life_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[list[dict]]:
    ...
@overload
def updatedb_life_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[list[dict]]:
    ...
def updatedb_life_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[list[dict]] | AsyncIterator[list[dict]]:
    """持续采集 115 生活日志，以更新 SQLite 数据库

    .. note::
        当 ``from_id < 0`` 时，会从数据库获取最大 id 作为 ``from_id``，获取不到时设为 0。
        当 ``from_id != 0`` 时，如果 from_time 为 0，则自动重设为 -1。

    :param client: 115 网盘客户端对象
    :param dbfile: 数据库文件路径，如果为 None，则自动确定
    :param from_id: 开始的事件 id （不含），若 < 0 则是从数据库获取最大 id
    :param from_time: 开始时间（含），若为 0 则从当前时间开始，若 < 0 则从最早开始
    :param cooldown: 冷却时间，大于 0 时，两次接口调用之间至少间隔这么多秒
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，每次产生一批事件（从当前到上次截止）

    .. code::

        from time import sleep
        from p115client import P115Client
        from p115client.tool import updatedb_life_iter

        client = P115Client.from_path()

        for event_list in updatedb_life_iter(client):
            if event_list:
                print("采集到操作事件列表:", event_list)
            else:
                sleep(1)
    """
    client, con = _init_client(client, dbfile)
    def gen_step():
        nonlocal from_id
        if from_id < 0:
            from_id = yield wrap_async(find, async_, threaded=True)(
                con, 
                "SELECT MAX(id) FROM life", 
                default=0, 
            )
        with with_iter_next(iter_life_behavior_list(
            client, 
            from_id=from_id, 
            from_time=from_time, 
            ignore_types=(10,), 
            cooldown=cooldown,         
            app=app, 
            async_=async_, 
            **request_kwargs, 
        )) as get_next:
            while True:
                event_list = yield get_next()
                event_list.reverse()
                if attrs := list(map(event_normalize_attr, event_list)):
                    if news := [a for a in attrs if a["is_alive"]]:
                        attrs.extend((yield load_missing_ancestors(
                            client, 
                            con, 
                            news, 
                            cooldown=cooldown, 
                            app=app, 
                            async_=async_, 
                            **request_kwargs, 
                        )))
                    yield wrap_async(upsert_items, async_, threaded=True)(
                        con, attrs, commit=True)
                if event_list:
                    yield wrap_async(execute, async_, threaded=True)(
                        con, 
                        "INSERT OR IGNORE INTO life(id, data) VALUES (?, ?)", 
                        [(int(event["id"]), dumps(event)) for event in event_list], 
                        commit=True, 
                    )
                yield Yield(event_list)
    return run_gen_step_iter(gen_step, async_)


@overload
def updatedb_history_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[list[dict]]:
    ...
@overload
def updatedb_history_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[list[dict]]:
    ...
def updatedb_history_iter(
    client: str | PathLike | P115Client, 
    dbfile: None | str | PathLike | Connection | Cursor = None, 
    from_id: int = -1, 
    from_time: float = 0, 
    cooldown: float = 0.2, 
    app: str = "android", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[list[dict]] | AsyncIterator[list[dict]]:
    """持续采集 115 历史记录，以更新 SQLite 数据库

    .. note::
        当 ``from_id < 0`` 时，会从数据库获取最大 id 作为 ``from_id``，获取不到时设为 0。
        当 ``from_id != 0`` 时，如果 from_time 为 0，则自动重设为 -1。

    :param client: 115 网盘客户端对象
    :param dbfile: 数据库文件路径，如果为 None，则自动确定
    :param from_id: 开始的事件 id （不含），若 < 0 则是从数据库获取最大 id
    :param from_time: 开始时间（含），若为 0 则从当前时间开始，若 < 0 则从最早开始
    :param cooldown: 冷却时间，大于 0 时，两次接口调用之间至少间隔这么多秒
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，每次产生一批事件（从当前到上次截止）

    .. code::

        from time import sleep
        from p115client import P115Client
        from p115client.tool import updatedb_history_iter

        client = P115Client.from_path()

        for event_list in updatedb_history_iter(client):
            if event_list:
                print("采集到历史记录列表:", event_list)
            else:
                sleep(1)
    """
    client, con = _init_client(client, dbfile)
    def gen_step():
        nonlocal from_id
        if from_id < 0:
            from_id = yield wrap_async(find, async_, threaded=True)(
                con, 
                "SELECT MAX(id) FROM history", 
                default=0, 
            )
        with with_iter_next(iter_history_list(
            client, 
            from_id=from_id, 
            from_time=from_time, 
            ignore_types=(), 
            cooldown=cooldown,         
            app=app, 
            async_=async_, 
            **request_kwargs, 
        )) as get_next:
            while True:
                event_list = yield get_next()
                event_list.reverse()
                if attrs := list(map(event_normalize_attr, event_list)):
                    if news := [a for a in attrs if a["is_alive"]]:
                        attrs.extend((yield load_missing_ancestors(
                            client, 
                            con, 
                            news, 
                            cooldown=cooldown, 
                            app=app, 
                            async_=async_, 
                            **request_kwargs, 
                        )))
                    yield wrap_async(upsert_items, async_, threaded=True)(
                        con, attrs, commit=True)
                if event_list:
                    yield wrap_async(execute, async_, threaded=True)(
                        con, 
                        "INSERT OR IGNORE INTO history(id, data) VALUES (?, ?)", 
                        [(int(event["id"]), dumps(event)) for event in event_list], 
                        commit=True, 
                    )
                yield Yield(event_list)
    return run_gen_step_iter(gen_step, async_)


class P115QueryDB:
    """封装了一些常用的数据库查询方法，针对 updatedb 产生的 SQLite 数据库

    .. note::
        默认情况下，只有 "id"、"parent_id"、"updated_at" 有索引，所以请自行添加其它需要的索引
    """
    __slots__ = "con",

    def __init__(self, con, /):
        self.con = connect(con, check_same_thread=False, timeout=inf)

    def get_ancestors(
        self, 
        id: int = 0, 
        /, 
    ) -> list[dict]:
        """获取某个文件或目录的祖先节点信息，包括 "id"、"parent_id" 和 "name" 字段

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id

        :return: 当前节点的祖先节点列表，从根目录开始（id 为 0）直到当前节点
        """
        ancestors = [{"id": 0, "parent_id": 0, "name": ""}]
        if not id:
            return ancestors
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        ls = list(query(con, """\
WITH t AS (
    SELECT id, parent_id, name FROM data WHERE id = ?
    UNION ALL
    SELECT data.id, data.parent_id, data.name FROM t JOIN data ON (t.parent_id = data.id)
)
SELECT id, parent_id, name FROM t;""", id))
        if not ls:
            raise FileNotFoundError(errno.ENOENT, id)
        if ls[-1][1]:
            raise ValueError(f"dangling id: {id}")
        ancestors.extend(dict(zip(("id", "parent_id", "name"), record)) for record in reversed(ls))
        return ancestors

    def get_attr(
        self, 
        id: int = 0, 
        /, 
    ) -> dict:
        """获取某个文件或目录的信息

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id

        :return: 当前节点的信息字典
        """
        if not id:
            return {
                "id": 0, "parent_id": 0, "name": "", "sha1": "", "size": 0, "pickcode": "", 
                "is_dir": 1, "is_alive": 1, "extra": None, "created_at": 0, "updated_at": 0, 
            }
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        return find(
            con, 
            f"SELECT * FROM data WHERE id=? LIMIT 1", 
            id, 
            FileNotFoundError(errno.ENOENT, id), 
            row_factory="dict", 
        )

    def get_id(
        self, 
        /, 
        pickcode: str = "", 
        sha1: str = "", 
        path: str | Sequence[str] = "", 
        parent_id:  int = 0, 
        is_alive: bool = True, 
    ) -> int:
        """查询匹配某个字段的文件或目录的 id

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param pickcode: 当前节点的提取码，优先级高于 sha1
        :param sha1: 当前节点的 sha1 校验哈希值，优先级高于 path
        :param path: 当前节点的路径
        :param parent_id: 仅用于 `path` 参数，用来限定搜索的顶层路径

        :return: 当前节点的 id
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        insertion = " AND is_alive" if is_alive else ""
        if pickcode:
            return find(
                con, 
                f"SELECT id FROM data WHERE pickcode=?{insertion} LIMIT 1", 
                pickcode, 
                default=FileNotFoundError(pickcode), 
            )
        elif sha1:
            return find(
                con, 
                f"SELECT id FROM data WHERE sha1=?{insertion} LIMIT 1", 
                sha1, 
                default=FileNotFoundError(sha1), 
            )
        elif path:
            return P115QueryDB.path_to_id(
                con, 
                path, 
                parent_id=parent_id, 
                is_alive=is_alive, 
            )
        return 0

    def get_parent_id(
        self, 
        id: int = 0, 
        /, 
        default: None | int = None, 
    ) -> int:
        """获取某个 id 对应的父 id

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id

        :return: 当前节点的父 id
        """
        if id == 0:
            return 0
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = "SELECT parent_id FROM data WHERE id=?"
        return find(con, sql, id, FileNotFoundError(errno.ENOENT, id) if default is None else default)

    def get_path(
        self, 
        id: int = 0, 
        /, 
    ) -> str:
        """获取某个文件或目录的路径

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id

        :return: 当前节点的路径
        """
        if not id:
            return "/"
        ancestors = P115QueryDB.get_ancestors(self, id)
        return "/".join(escape(a["name"]) for a in ancestors)

    def get_patht(
        self, 
        id: int = 0, 
        /, 
    ) -> list[str]:
        """获取某个文件或目录的路径节点元组

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id

        :return: 当前节点的路径节点元组
        """
        if not id:
            return [""]
        ancestors = P115QueryDB.get_ancestors(self, id)
        return [a["name"] for a in ancestors]

    def get_pickcode(
        self, 
        /, 
        id: int = -1, 
        sha1: str = "", 
        path: str | Sequence[str] = "", 
        parent_id: int = 0, 
        is_alive: bool = True, 
    ) -> str:
        """查询匹配某个字段的文件或目录的提取码

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id，优先级高于 sha1
        :param sha1: 当前节点的 sha1 校验哈希值，优先级高于 path
        :param path: 当前节点的路径
        :param parent_id: 仅用于 `path` 参数，用来限定搜索的顶层路径

        :return: 当前节点的提取码
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        insertion = " AND is_alive" if is_alive else ""
        if id >= 0:
            if not id:
                return ""
            return find(
                con, 
                f"SELECT pickcode FROM data WHERE id=?{insertion} LIMIT 1;", 
                id, 
                default=FileNotFoundError(id), 
            )
        elif sha1:
            return find(
                con, 
                f"SELECT pickcode FROM data WHERE sha1=?{insertion} LIMIT 1;", 
                sha1, 
                default=FileNotFoundError(sha1), 
            )
        else:
            if path in ("", "/"):
                return ""
            id = P115QueryDB.path_to_id(con, path, parent_id=parent_id, is_alive=is_alive)
            return P115QueryDB.get_pickcode(con, id)

    def get_sha1(
        self, 
        /, 
        id: int = -1, 
        pickcode: str = "", 
        path: str | Sequence[str] = "", 
        parent_id: int = 0, 
        is_alive: bool = True, 
    ) -> str:
        """查询匹配某个字段的文件的 sha1

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id，优先级高于 pickcode
        :param pickcode: 当前节点的提取码，优先级高于 path
        :param path: 当前节点的路径
        :param parent_id: 仅用于 `path` 参数，用来限定搜索的顶层路径

        :return: 当前节点的 sha1 校验哈希值
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        insertion = " AND is_alive" if is_alive else ""
        if id >= 0:
            if not id:
                return ""
            return find(
                con, 
                f"SELECT sha1 FROM data WHERE id=?{insertion} LIMIT 1;", 
                id, 
                default=FileNotFoundError(id), 
            )
        elif pickcode:
            return find(
                con, 
                f"SELECT sha1 FROM data WHERE pickcode=?{insertion} LIMIT 1;", 
                pickcode, 
                default=FileNotFoundError(pickcode), 
            )
        else:
            if path in ("", "/"):
                return ""
            id = P115QueryDB.path_to_id(con, path, parent_id=parent_id, is_alive=is_alive)
            return P115QueryDB.get_sha1(con, id)

    def has_id(
        self, 
        id: int, 
        /, 
        is_alive: bool = True, 
    ) -> int:
        """是否存在某个 id

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param id: 当前节点的 id
        :param is_alive: 是否存活

        :return: 如果是 1，则是 True；如果是 0，则是 False
        """
        if id == 0:
            return 1
        elif id < 0:
            return 0
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = "SELECT 1 FROM data WHERE id=?"
        if is_alive:
            sql += " AND is_alive"
        return find(con, sql, id, 0)

    def path_to_id(
        self, 
        /, 
        path: str | Sequence[str] = "", 
        ensure_file: None | bool = None, 
        parent_id: int = 0, 
        is_alive: bool = True, 
    ) -> int:
        """查询匹配某个路径的文件或目录的信息字典，只返回找到的第 1 个

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param path: 路径
        :param ensure_file: 是否文件

            - 如果为 True，必须是文件
            - 如果为 False，必须是目录
            - 如果为 None，可以是文件或目录

        :param parent_id: 顶层目录的 id

        :return: 找到的第 1 个匹配的节点 id
        """
        try:
            return next(P115QueryDB.iter_id_to_path(
                self, 
                path=path, 
                ensure_file=ensure_file, 
                parent_id=parent_id, 
                is_alive=is_alive, 
            ))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, path) from None

    def iter_children(
        self, 
        parent_id: int = 0, 
        /, 
        fields: Collection[str] = (), 
        ensure_file: None | bool = None, 
    ) -> Iterator[dict]:
        """获取某个目录之下的文件或目录的信息

        .. caution::
            当 ``fields`` 为空时，获取全部字段

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param parent_id: 父目录的 id
        :param fields: 需要获取的字段
        :param ensure_file: 是否仅输出文件

            - 如果为 True，仅输出文件
            - 如果为 False，仅输出目录
            - 如果为 None，全部输出

        :return: 迭代器，产生一组信息的字典，大概包含如下字段（具体由你的数据库 ``data`` 表确定）：

            .. code:: python

                (
                    "id", "parent_id", "name", "sha1", "size", "pickcode", 
                    "is_dir", "is_alive", "extra", "created_at", "updated_at", 
                )
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        if fields:
            sql = f"SELECT {','.join(fields)} FROM data WHERE parent_id=? AND is_alive"
        else:
            sql = "SELECT * FROM data WHERE parent_id=? AND is_alive"
        if ensure_file is not None:
            if ensure_file:
                sql += " AND NOT is_dir"
            else:
                sql += " AND is_dir"
        return query(con, sql, parent_id, row_factory="dict")

    def iter_count_dir(
        self, 
        parent_id: int = 0, 
        /, 
    ) -> Iterator[dict]:
        """迭代获取所有指定 id 下所有目录节点（包括自己）直属的文件数和目录数

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param parent_id: 顶层目录的 id

        :return: 迭代器，返回字典

            .. code::

                {
                    "id": int, 
                    "parent_id": int, 
                    "dir_count": int, 
                    "file_count": int, 
                }
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = """\
WITH t AS (
    SELECT id, parent_id, is_dir FROM data WHERE parent_id = ? AND is_alive
    UNION ALL
    SELECT data.id, data.parent_id, data.is_dir FROM t JOIN data ON (t.id = data.parent_id) WHERE is_alive
), count AS (
    SELECT parent_id AS id, SUM(is_dir) AS dir_count, SUM(NOT is_dir) AS file_count 
    FROM t 
    GROUP BY parent_id
)
SELECT data.parent_id, count.* FROM count JOIN data USING (id)"""
        return query(con, sql, parent_id, row_factory="dict")

    def iter_count_tree(
        self, 
        parent_id: int = 0, 
        /, 
    ) -> Iterator[dict]:
        """迭代获取所有指定 id 下所有目录节点（包括自己）直属的文件数和目录数，以及子树下的文件数合计和目录数合计

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param parent_id: 顶层目录的 id

        :return: 迭代器，返回字典

            .. code::

                {
                    "id": int, 
                    "parent_id": int, 
                    "dir_count": int, 
                    "file_count": int, 
                    "tree_dir_count": int, 
                    "tree_file_count": int, 
                }
        """
        data = {a["id"]: a for a in P115QueryDB.iter_count_dir(self, parent_id)}
        id_to_children = grouped_mapping((a["parent_id"], id) for id, a in data.items())
        def calc(attr: dict, /) -> dict:
            if children := id_to_children.get(attr["id"]):
                for cid in children:
                    cattr = data[cid]
                    if "tree_dir_count" not in cattr:
                        calc(cattr)
                    attr["tree_dir_count"] = attr.get("tree_dir_count", 0) + 1 + cattr["tree_dir_count"]
                    attr["tree_file_count"] = attr.get("tree_file_count", attr["file_count"]) + cattr["tree_file_count"]
            elif "tree_dir_count" not in attr:
                attr["tree_dir_count"] = attr["dir_count"]
                attr["tree_file_count"] = attr["file_count"]
            return attr
        return map(calc, data.values())

    def iter_dangling_ids(
        self, 
        /, 
    ) -> Iterator[int]:
        """罗列所有悬空的文件或目录的 id

        .. note::
            悬空的 id，即祖先节点中，存在一个节点，它的 parent_id 是悬空的

        :param self: P115QueryDB 实例或者数据库连接或游标

        :return: 迭代器，一组目录的 id
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = """\
WITH dangling_ids(id) AS (
    SELECT d1.id
    FROM data AS d1 LEFT JOIN data AS d2 ON (d1.parent_id = d2.id)
    WHERE d1.parent_id AND d2.id IS NULL
    UNION ALL
    SELECT data.id FROM dangling_ids JOIN data ON (dangling_ids.id = data.parent_id)
)
SELECT id FROM dangling_ids"""
        return query(con, sql, row_factory="one")

    def iter_dangling_parent_ids(
        self, 
        /, 
    ) -> Iterator[int]:
        """罗列所有悬空的 parent_id

        .. note::
            悬空的 parent_id，即所有的 parent_id 中，，不为 0 且不在 `data` 表中的部分

        :param self: P115QueryDB 实例或者数据库连接或游标

        :return: 迭代器，一组目录的 id
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = """\
SELECT
    DISTINCT d1.parent_id
FROM
    data AS d1 LEFT JOIN data AS d2 ON (d1.parent_id = d2.id)
WHERE
    d1.parent_id AND d2.id IS NULL"""
        return query(con, sql, row_factory="one")

    @overload
    def iter_descendants(
        self, 
        parent_id: int = 0, 
        /, 
        min_depth: int = 1, 
        max_depth: int = -1, 
        *, 
        fields: str, 
        escape: None | bool | Callable[[str], str] = True, 
        ensure_file: None | bool = None, 
        topdown: None | bool = True, 
    ) -> Iterator[Any]:
        ...
    @overload
    def iter_descendants(
        self, 
        parent_id: int = 0, 
        /, 
        min_depth: int = 1, 
        max_depth: int = -1, 
        *, 
        fields: Collection[str] | bool = True, 
        escape: None | bool | Callable[[str], str] = True, 
        ensure_file: None | bool = None, 
        topdown: None | bool = True, 
    ) -> Iterator[dict]:
        ...
    def iter_descendants(
        self, 
        parent_id: int = 0, 
        /, 
        min_depth: int = 1, 
        max_depth: int = -1, 
        *, 
        fields: str | Collection[str] | bool = True, 
        escape: None | bool | Callable[[str], str] = True, 
        ensure_file: None | bool = None, 
        topdown: None | bool = True, 
    ) -> Iterator:
        """获取某个目录之下的所有节点信息

        .. caution::
            当 ``fields`` 为空时，获取全部字段

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param parent_id: 顶层目录的 id
        :param min_depth: 最小深度
        :param max_depth: 最大深度。如果小于 0，则无限深度
        :param fields: 需要获取的字段

            - 如果为 str，直接获取这个字段的值（不返回字典）
            - 如果为 True，获取所有字段，且包括（"ancestors", "path", "relpath", "depth"），返回字典
            - 如果为 False，获取所有字段，但除了（"ancestors", "path", "relpath", "depth"），返回字典
            - 否则，获取所指定的这组字段，返回字典

        :param escape: 对文件名进行转义

            - 如果为 None，则不处理；否则，这个函数用来对文件名中某些符号进行转义，例如 "/" 等
            - 如果为 True，则使用 `posixpatht.escape`，会对文件名中 "/"，或单独出现的 "." 和 ".." 用 "\\" 进行转义
            - 如果为 False，则使用 `posix_escape_name` 函数对名字进行转义，会把文件名中的 "/" 转换为 "|"
            - 如果为 Callable，则用你所提供的调用，以或者转义后的名字

        :param ensure_file: 是否仅输出文件

            - 如果为 True，仅输出文件
            - 如果为 False，仅输出目录
            - 如果为 None，全部输出

        :param topdown: 是否自顶向下深度优先遍历

            - 如果为 True，则自顶向下深度优先遍历
            - 如果为 False，则自底向上深度优先遍历
            - 如果为 None，则自顶向下宽度优先遍历

        :return: 迭代器，产生一组信息的字典，大概包含如下字段：

            .. code:: python

                (
                    # NOTE: 这些具体由你的数据库 ``data`` 表确定
                    "id", "parent_id", "name", "sha1", "size", "pickcode", 
                    "is_dir", "is_alive", "extra", "created_at", "updated_at", 
                    # NOTE: 这些是另外附加的字段
                    "ancestors", "path", "relpath", "depth", 
                )
        """
        if 0 <= max_depth < min_depth:
            return
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        if isinstance(escape, bool):
            if escape:
                from posixpatht import escape
            else:
                escape = posix_escape_name
        escape = cast(None | Callable[[str], str], escape)
        field: str = ""
        if isinstance(fields, bool) or not fields:
            with_ancestors = with_path = with_relpath = with_depth = fields if isinstance(fields, bool) else True
            children_fields = set()
        else:
            if isinstance(fields, str):
                field = fields
                fields = {field}
            else:
                fields = set(fields)
            with_ancestors = "ancestors" in fields
            with_path = "path" in fields
            with_relpath = "relpath" in fields
            with_depth = "depth" in fields
            children_fields = fields - frozenset(("ancestors", "path", "relpath", "depth")) | {"id", "is_dir"}
            if with_ancestors or with_path or with_relpath:
                children_fields |= frozenset(("name", "parent_id"))
        ancestors: list[dict] = []
        dir_: str = ""
        reldir: str = ""
        if parent_id:
            if with_ancestors or with_path:
                ancestors = P115QueryDB.get_ancestors(con, parent_id)
                if with_path:
                    if escape is None:
                        dir_ = "".join(a["name"] + "/" for a in ancestors)
                    else:
                        dir_ = "".join(escape(a["name"]) + "/" for a in ancestors)
        else:
            if with_ancestors:
                ancestors = [{"id": 0, "parent_id": 0, "name": ""}]
            if with_path:
                dir_ = "/"
        def may_yield(attr: dict, /):
            if ensure_file is None or (attr["is_dir"] ^ ensure_file):
                if field:
                    yield attr[field]
                else:
                    yield attr
        children_ensure_file = False if ensure_file is False else None
        if topdown is None:
            gen = bfs_gen((parent_id, 0, ancestors, dir_, reldir))
            send: Callable = gen.send
            for parent_id, depth, ancestors, dir_, reldir in gen:
                depth += 1
                will_step_in = max_depth < 0 or depth < max_depth
                will_yield = min_depth <= depth and (max_depth < 0 or depth <= max_depth)
                for attr in P115QueryDB.iter_children(
                    con, 
                    parent_id, 
                    fields=children_fields, 
                    ensure_file=children_ensure_file, 
                ):
                    if with_depth:
                        attr["depth"] = depth
                    if with_ancestors:
                        attr["ancestors"] = [
                            *ancestors, 
                            {k: attr[k] for k in ("id", "parent_id", "name")}, 
                        ]
                    if with_path or with_relpath:
                        name = attr["name"]
                        if escape is not None:
                            name = escape(name)
                        if with_path:
                            attr["path"] = dir_ + name
                        if with_relpath:
                            attr["relpath"] = reldir + name
                    if will_step_in and attr["is_dir"]:
                        send((
                            attr["id"], 
                            depth, 
                            attr["ancestors"] if with_ancestors else None, 
                            attr["path"] + "/" if with_path else "", 
                            attr["relpath"] + "/" if with_relpath else "", 
                        ))
                    if will_yield:
                        yield from may_yield(attr)
        else:
            cache: dict[Iterator, dict] = {}
            stack: list[tuple[Iterator[dict], list[dict], str, str]] = [(
                iter(tuple(P115QueryDB.iter_children(
                    con, 
                    parent_id, 
                    fields=children_fields, 
                    ensure_file=children_ensure_file, 
                ))), ancestors, dir_, reldir)]
            depth = 0
            while depth >= 0:
                attrs, ancestors, dir_, reldir = stack[depth]
                depth += 1
                will_step_in = max_depth < 0 or depth < max_depth
                will_yield = min_depth <= depth and (max_depth < 0 or depth <= max_depth)
                for attr in attrs:
                    if with_depth:
                        attr["depth"] = depth
                    if with_ancestors:
                        attr["ancestors"] = [
                            *ancestors, 
                            {k: attr[k] for k in ("id", "parent_id", "name")}, 
                        ]
                    if with_path or with_relpath:
                        name = attr["name"]
                        if escape is not None:
                            name = escape(name)
                        if with_path:
                            attr["path"] = dir_ + name
                        if with_relpath:
                            attr["relpath"] = reldir + name
                    if will_yield and topdown:
                        yield from may_yield(attr)
                    if will_step_in and attr["is_dir"]:
                        attrs = iter(tuple(P115QueryDB.iter_children(
                            con, 
                            attr["id"], 
                            fields=children_fields, 
                            ensure_file=children_ensure_file, 
                        )))
                        quadruple = (
                            attrs, 
                            attr["ancestors"] if with_ancestors else ancestors, 
                            attr["path"] + "/" if with_path else "", 
                            attr["relpath"] + "/" if with_relpath else "", 
                        )
                        try:
                            stack[depth] = quadruple
                        except IndexError:
                            stack.append(quadruple)
                        if will_yield and not topdown:
                            cache[attrs] = attr
                        break
                    if will_yield and not topdown:
                        yield from may_yield(attr)
                else:
                    if cache and attrs in cache:
                        yield from may_yield(cache.pop(attrs))
                    depth -= 2

    @overload
    def iter_descendants_bfs(
        self, 
        parent_id: int = 0, 
        /, 
        *, 
        fields: str, 
        escape: None | bool | Callable[[str], str] = True, 
    ) -> Iterator[Any]:
        ...
    @overload
    def iter_descendants_bfs(
        self, 
        parent_id: int = 0, 
        /, 
        *, 
        fields: Collection[str] | bool = True, 
        escape: None | bool | Callable[[str], str] = True, 
    ) -> Iterator[dict]:
        ...
    def iter_descendants_bfs(
        self, 
        parent_id: int = 0, 
        /, 
        *, 
        fields: str | Collection[str] | bool = True, 
        escape: None | bool | Callable[[str], str] = True, 
    ) -> Iterator:
        """获取某个目录之下的所有节点信息

        .. caution::
            当 ``fields`` 为空时，获取全部字段        

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param parent_id: 顶层目录的 id
        :param fields: 需要获取的字段

            - 如果为 str，直接获取这个字段的值（不返回字典）
            - 如果为 True，获取所有字段，且包括（"ancestors", "path", "relpath", "depth"），返回字典
            - 如果为 False，获取所有字段，但除了（"ancestors", "path", "relpath", "depth"），返回字典
            - 否则，获取所指定的这组字段，返回字典

        :param escape: 对文件名进行转义

            - 如果为 None，则不处理；否则，这个函数用来对文件名中某些符号进行转义，例如 "/" 等
            - 如果为 True，则使用 `posixpatht.escape`，会对文件名中 "/"，或单独出现的 "." 和 ".." 用 "\\" 进行转义
            - 如果为 False，则使用 `posix_escape_name` 函数对名字进行转义，会把文件名中的 "/" 转换为 "|"
            - 如果为 Callable，则用你所提供的调用，以或者转义后的名字

        :return: 迭代器，产生一组信息的字典，大概包含如下字段：

            .. code:: python

                (
                    # NOTE: 这些具体由你的数据库 ``data`` 表确定
                    "id", "parent_id", "name", "sha1", "size", "pickcode", 
                    "is_dir", "is_alive", "extra", "created_at", "updated_at", 
                    # NOTE: 这些是另外附加的字段
                    "ancestors", "path", "relpath", "depth", 
                )
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        if isinstance(escape, bool):
            if escape:
                from posixpatht import escape
            else:
                escape = posix_escape_name
        escape = cast(None | Callable[[str], str], escape)
        field: str = ""
        if isinstance(fields, bool) or not fields:
            with_depth = with_ancestors = with_path = with_relpath = fields if isinstance(fields, bool) else True
            fields = "*"
            fields1 = "*"
            fields2 = "data.*"
        else:
            if isinstance(fields, str):
                field = fields
                fields = {field}
            else:
                fields = set(fields)
            fields.add("id")
            with_depth = "depth" in fields
            with_ancestors = "ancestors" in fields
            with_path = "path" in fields
            with_relpath = "relpath" in fields
            if with_depth or with_ancestors or with_path or with_relpath:
                fields.add("parent_id")
                fields.add("name")
            fields -= frozenset(("depth", "ancestors", "path", "relpath"))
            fields1 = ",".join(fields)
            fields2 = ",".join("data." + f for f in fields)
        sql = f"""\
WITH t AS (
    SELECT {fields1} FROM data WHERE parent_id={parent_id:d} AND is_alive
    UNION ALL
    SELECT {fields2} FROM t JOIN data ON(t.id = data.parent_id) WHERE data.is_alive
) SELECT * FROM t"""
        row_factory: str | Callable = "any"
        if with_depth or with_ancestors or with_path or with_relpath:
            if with_depth:
                d_depth = {parent_id: 0}
            if with_ancestors or with_path:
                if parent_id:
                    ancestors = P115QueryDB.get_ancestors(con, parent_id)
                    if with_ancestors:
                        d_ancestors = {parent_id: ancestors}
                    if with_path:
                        if escape is None:
                            d_path = {parent_id: "".join(a["name"] + "/" for a in ancestors)}
                        else:
                            d_path = {parent_id: "".join(escape(a["name"]) + "/" for a in ancestors)}
                else:
                    if with_ancestors:
                        d_ancestors = {0: [{"id": 0, "parent_id": 0, "name": ""}]}
                    if with_path:
                        d_path = {0: "/"}
            if with_relpath:
                d_relpath = {parent_id: ""}
            cursor_fields: None | tuple[str, ...] = None
            def row_factory(cursor, record, /):
                nonlocal cursor_fields
                if cursor_fields is None:
                    cursor_fields = tuple(f[0] for f in cursor.description)
                attr = dict(zip(cursor_fields, record))
                id, pid, name = attr["id"], attr["parent_id"], attr["name"]
                if with_depth:
                    d_depth[id] = attr["depth"] = d_depth[pid] + 1
                if with_ancestors:
                    attr["ancestors"] = [*d_ancestors[pid], {"id": id, "parent_id": pid, "name": name}]
                if escape is not None and (with_path or with_relpath):
                    name = escape(name)
                if with_path:
                    attr["path"] = d_path[pid] + name
                if with_relpath:
                    attr["relpath"] = d_relpath[pid] + name
                if attr.get("is_dir", True):
                    if with_ancestors:
                        d_ancestors[id] = attr["ancestors"]
                    if with_path:
                        d_path[id] = attr["path"] + "/"
                    if with_relpath:
                        d_relpath[id] = attr["relpath"] + "/"
                if field:
                    return attr[field]
                return attr
        elif field:
            row_factory = "one"
        else:
            row_factory = "dict"
        return query(con, sql, row_factory=row_factory)

    def iter_dup_files(
        self, 
        /, 
    ) -> Iterator[dict]:
        """罗列所有重复文件

        :param self: P115QueryDB 实例或者数据库连接或游标

        :return: 迭代器，一组文件的信息
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = f"""\
WITH stats AS (
    SELECT
        COUNT(1) OVER w AS total, 
        ROW_NUMBER() OVER w AS nth, 
        *
    FROM data
    WHERE NOT is_dir AND is_alive
    WINDOW w AS (PARTITION BY sha1, size)
)
SELECT * FROM stats WHERE total > 1"""
        return query(con, sql, row_factory="dict")

    def iter_existing_id(
        self, 
        ids: Iterable[int], 
        /, 
        is_alive: bool = True, 
    ) -> Iterator[int]:
        """筛选出一系列 id 中，在数据库中存在的

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param ids: 一系列的节点 id
        :param is_alive: 是否存活

        :return: 一系列 id 中在数据库中存在的那些的迭代器（实际直接返回一个游标）
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = "SELECT id FROM data WHERE id IN (%s)" % (",".join(map("%d".__mod__, ids)) or "NULL")
        if is_alive:
            sql += " AND is_alive"
        return query(con, sql, row_factory="one")

    def iter_id_to_parent_id(
        self, 
        ids: Iterable[int], 
        /, 
        recursive: bool = False, 
    ) -> Iterator[tuple[int, int]]:
        """找出一系列 id 所对应的父 id，返回 ``(id, parent_id)`` 的 2 元组

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param ids: 一系列的节点 id
        :param recursive: 是否递归，如果为 True，则还会处理它们的祖先节点

        :return: 迭代器，产生 ``(id, parent_id)`` 的 2 元组（实际直接返回一个游标）
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        s_ids = "(%s)" % (",".join(map(str, ids)) or "NULL")
        if recursive:
            sql = """\
WITH pairs AS (
    SELECT id, parent_id FROM data WHERE id IN %s
    UNION ALL
    SELECT data.id, data.parent_id FROM pairs JOIN data ON (pairs.parent_id = data.id)
) SELECT * FROM pairs""" % s_ids
        else:
            sql = "SELECT id, parent_id FROM data WHERE id IN %s" % s_ids
        return query(con, sql)

    def iter_id_to_path(
        self, 
        /, 
        path: str | Sequence[str] = "", 
        ensure_file: None | bool = None, 
        parent_id: int = 0, 
        is_alive: bool = True, 
    ) -> Iterator[int]:
        """查询匹配某个路径的文件或目录的信息字典

        .. note::
            同一个路径可以有多条对应的数据

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param path: 路径
        :param ensure_file: 是否文件

            - 如果为 True，必须是文件
            - 如果为 False，必须是目录
            - 如果为 None，可以是文件或目录

        :param parent_id: 顶层目录的 id

        :return: 迭代器，产生一组匹配指定路径的（文件或目录）节点的 id
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        patht: Sequence[str]
        if isinstance(path, str):
            if ensure_file is None and path_is_dir_form(path):
                ensure_file = False
            patht, _ = splits("/" + path)
        else:
            patht = ("", *filter(None, path))
        if not parent_id and len(patht) == 1:
            return iter((0,))
        insertion = " AND is_alive" if is_alive else ""
        if len(patht) > 2:
            sql = f"SELECT id FROM data WHERE parent_id=? AND name=?{insertion} AND is_dir LIMIT 1"
            for name in patht[1:-1]:
                parent_id = find(con, sql, (parent_id, name), default=-1)
                if parent_id < 0:
                    return iter(())
        sql = f"SELECT id FROM data WHERE parent_id=? AND name=?{insertion}"
        if ensure_file is None:
            sql += " ORDER BY is_dir DESC"
        elif ensure_file:
            sql += " AND NOT is_dir"
        else:
            sql += " AND is_dir LIMIT 1"
        return query(con, sql, (parent_id, patht[-1]), row_factory="one")

    def iter_parent_id(
        self, 
        ids: Iterable[int], 
        /, 
    ) -> Iterator[int]:
        """找出一系列 id 所对应的父 id

        :param self: P115QueryDB 实例或者数据库连接或游标
        :param ids: 一系列的节点 id

        :return: 它们的父 id 的迭代器（实际直接返回一个游标）
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        sql = "SELECT parent_id FROM data WHERE id IN (%s)" % (",".join(map("%d".__mod__, ids)) or "NULL")
        return query(con, sql, row_factory="one")

    def select_na_ids(
        self, 
        /, 
    ) -> set[int]:
        """找出所有的失效节点和悬空节点的 id

        .. note::
            悬空节点，就是此节点有一个祖先节点的 parant_id，不为 0 且不在 `data` 表中

        :param self: P115QueryDB 实例或者数据库连接或游标

        :return: 一组悬空节点的 id 的集合
        """
        con: Any
        if isinstance(self, P115QueryDB):
            con = self.con
        else:
            con = self
        ok_ids: set[int] = set(query(con, "SELECT id FROM data WHERE NOT is_alive", row_factory="one"))
        na_ids: set[int] = set()
        d = dict(query(con, "SELECT id, parent_id FROM data WHERE is_alive"))
        temp: list[int] = []
        push = temp.append
        clear = temp.clear
        update_ok = ok_ids.update
        update_na = na_ids.update
        for k in d:
            try:
                push(k)
                while k := d[k]:
                    if k in ok_ids:
                        update_ok(temp)
                        break
                    elif k in na_ids:
                        update_na(temp)
                        break
                    push(k)
                else:
                    update_ok(temp)
            except KeyError:
                update_na(temp)
            finally:
                clear()
        return na_ids




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "parse_export_dir_as_dict_iter", "parse_export_dir_as_path_iter", 
    "parse_export_dir_as_patht_iter", "export_dir", "export_dir_result", 
    "export_dir_parse_iter", 
]
__doc__ = "这个模块提供了一些和导出目录树有关的函数"

from asyncio import sleep as async_sleep
from collections.abc import (
    AsyncIterable, AsyncIterator, Callable, Coroutine, Iterable, Iterator, 
)
from functools import partial
from io import BufferedReader, TextIOWrapper
from itertools import count
from os import PathLike
from re import compile as re_compile
from time import sleep, time
from typing import cast, overload, Any, Final, IO, Literal

from asynctools import ensure_async, ensure_aiter
from filewrap import AsyncBufferedReader, AsyncTextIOWrapper
from iterutils import (
    backgroud_loop, context, run_gen_step, run_gen_step_iter, 
    Yield, YieldFrom, 
)
from p115pickcode import to_id

from ..client import check_response, P115Client


CRE_TREE_PREFIX_match: Final = re_compile(r"^(?:\| )+\|-(.*)").match


@overload
def parse_export_dir_as_dict_iter(
    file: bytes | str | PathLike | Iterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[False] = False, 
) -> Iterator[dict]:
    ...
@overload
def parse_export_dir_as_dict_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[True], 
) -> AsyncIterator[dict]:
    ...
def parse_export_dir_as_dict_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[False, True] = False, 
) -> Iterator[dict] | AsyncIterator[dict]:
    """解析 115 导出的目录树（可通过 P115Client.fs_export_dir 提交导出任务）

    :param file: 文件路径、打开的文件或者迭代器（每次返回一行）
    :param encoding: 字符编码，对字节数据使用，转换为字符串
    :param close_file: 结束（包括异常退出）时尝试关闭 `file`
    :param async_: 是否异步

    :return: 把每一行解析为一个字典，迭代返回，格式为

        .. code:: python

            {
                "key":        int, # 序号
                "parent_key": int, # 上级目录的序号
                "depth":      int, # 深度
                "name":       str, # 名字
            }
    """
    if isinstance(file, (bytes, str, PathLike)):
        file = open(file, encoding=encoding)
        close_file = True
    def gen_step():
        it = ensure_aiter(file, threaded=True) if async_ else file
        do_next: Callable = anext if async_ else next # type: ignore
        stack: list[dict] = [{"key": 0, "parent_key": 0, "depth": 0, "name": ""}]
        push = stack.append
        root = yield do_next(it, None)
        if not root:
            return
        try:
            depth = 0
            for i in count(1):
                line = yield do_next(it)
                if not isinstance(line, str):
                    line = str(line, encoding)
                m = CRE_TREE_PREFIX_match(line)
                if m is None:
                    stack[depth]["name"] += "\n" + line[:-1]
                    continue
                else:
                    yield Yield(stack[depth])
                name = m[1]
                depth = (len(line) - len(name)) // 2 - 1
                item = {
                    "key": i, 
                    "parent_key": stack[depth-1]["key"], 
                    "depth": depth, 
                    "name": name, 
                }
                try:
                    stack[depth] = item
                except IndexError:
                    push(item)
        except (StopIteration, StopAsyncIteration):
            if depth:
                yield Yield(stack[depth])
        finally:
            if close_file:
                if async_:
                    if callable(aclose := getattr(file, "aclose", None)):
                        yield aclose()
                    elif callable(close := getattr(file, "close", None)):
                        yield ensure_async(close, threaded=True)
                elif callable(close := getattr(file, "close", None)):
                    close()
    return run_gen_step_iter(gen_step, async_)


@overload
def parse_export_dir_as_path_iter(
    file: bytes | str | PathLike | Iterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    escape: None | bool | Callable[[str], str] = True, 
    close_file: bool = False, 
    *, 
    async_: Literal[False] = False, 
) -> Iterator[str]:
    ...
@overload
def parse_export_dir_as_path_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    escape: None | bool | Callable[[str], str] = True, 
    close_file: bool = False, 
    *, 
    async_: Literal[True], 
) -> AsyncIterator[str]:
    ...
def parse_export_dir_as_path_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    escape: None | bool | Callable[[str], str] = True, 
    close_file: bool = False, 
    *, 
    async_: Literal[False, True] = False, 
) -> Iterator[str] | AsyncIterator[str]:
    """解析 115 导出的目录树（可通过 P115Client.fs_export_dir 提交导出任务）

    :param file: 文件路径、打开的文件或者迭代器（每次返回一行）
    :param encoding: 字符编码，对字节数据使用，转换为字符串
    :param escape: 对文件名进行转义

        - 如果为 None，则不处理；否则，这个函数用来对文件名中某些符号进行转义，例如 "/" 等
        - 如果为 True，则使用 `posixpatht.escape`，会对文件名中 "/"，或单独出现的 "." 和 ".." 用 "\\" 进行转义
        - 如果为 False，则使用 `posix_escape_name` 函数对名字进行转义，会把文件名中的 "/" 转换为 "|"
        - 如果为 Callable，则用你所提供的调用，以或者转义后的名字

    :param close_file: 结束（包括异常退出）时尝试关闭 `file`
    :param async_: 是否异步

    :return: 把每一行解析为一个路径，并逐次迭代返回
    """
    if isinstance(file, (bytes, str, PathLike)):
        file = open(file, encoding=encoding)
        close_file = True
    if isinstance(escape, bool):
        if escape:
            from posixpatht import escape
        else:
            from p115client.util import posix_escape_name as escape
    escape = cast(None | Callable[[str], str], escape)
    def gen_step():
        it = ensure_aiter(file, threaded=True) if async_ else file
        do_next: Callable = anext if async_ else next # type: ignore
        root = yield do_next(it, None)
        if not root:
            return
        if not isinstance(root, str):
            root = str(root, encoding)
        root = root.removesuffix("\n")[3:]
        if root == "根目录":
            stack = [""]
            root = "/"
        else:
            if escape is not None:
                root = escape(root)
            root = "/" + root
            stack = [root]
        push = stack.append
        try:
            depth = 0
            while True:
                line = yield do_next(it)
                if not isinstance(line, str):
                    line = str(line, encoding)
                m = CRE_TREE_PREFIX_match(line)
                if m is None:
                    stack[depth] += "\n" + line[:-1]
                    continue
                elif depth:
                    yield Yield(stack[depth])
                else:
                    yield Yield(root)
                name = m[1]
                depth = (len(line) - len(name)) // 2 - 1
                if escape is not None:
                    name = escape(name)
                path = stack[depth-1] + "/" + name
                try:
                    stack[depth] = path
                except IndexError:
                    push(path)
        except (StopIteration, StopAsyncIteration):
            if depth:
                yield Yield(stack[depth])
        finally:
            if close_file:
                if async_:
                    if callable(aclose := getattr(file, "aclose", None)):
                        yield aclose()
                    elif callable(close := getattr(file, "close", None)):
                        yield ensure_async(close, threaded=True)
                elif callable(close := getattr(file, "close", None)):
                    close()
    return run_gen_step_iter(gen_step, async_)


@overload
def parse_export_dir_as_patht_iter(
    file: bytes | str | PathLike | Iterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[False] = False, 
) -> Iterator[list[str]]:
    ...
@overload
def parse_export_dir_as_patht_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[True], 
) -> AsyncIterator[list[str]]:
    ...
def parse_export_dir_as_patht_iter(
    file: bytes | str | PathLike | Iterable[bytes | str] | AsyncIterable[bytes | str], 
    /, 
    encoding: str = "utf-16", 
    close_file: bool = False, 
    *, 
    async_: Literal[False, True] = False, 
) -> Iterator[list[str]] | AsyncIterator[list[str]]:
    """解析 115 导出的目录树（可通过 P115Client.fs_export_dir 提交导出任务）

    :param file: 文件路径、打开的文件或者迭代器（每次返回一行）
    :param encoding: 字符编码，对字节数据使用，转换为字符串
    :param close_file: 结束（包括异常退出）时尝试关闭 `file`
    :param async_: 是否异步

    :return: 把每一行解析为一个名字列表，并逐次迭代返回
    """
    if isinstance(file, (bytes, str, PathLike)):
        file = open(file, encoding=encoding)
        close_file = True
    def gen_step():
        it = ensure_aiter(file, threaded=True) if async_ else file
        do_next: Callable = anext if async_ else next # type: ignore
        root = yield do_next(it, None)
        if not root:
            return
        if not isinstance(root, str):
            root = str(root, encoding)
        root = root.removesuffix("\n")[3:]
        stack = [""]
        from_top_root = root == "根目录"
        if not from_top_root:
            stack.append(root)
        push = stack.append
        try:
            depth = len(stack) - 1
            while True:
                line = yield do_next(it)
                if not isinstance(line, str):
                    line = str(line, encoding)
                m = CRE_TREE_PREFIX_match(line)
                if m is None:
                    stack[depth] += "\n" + line[:-1]
                    continue
                else:
                    yield Yield(stack[:depth+1])
                name = m[1]
                depth = (len(line) - len(name)) // 2 - from_top_root
                try:
                    stack[depth] = name
                except IndexError:
                    push(name)
        except (StopIteration, StopAsyncIteration):
            if depth:
                yield Yield(stack[:depth+1])
        finally:
            if close_file:
                if async_:
                    if callable(aclose := getattr(file, "aclose", None)):
                        yield aclose()
                    elif callable(close := getattr(file, "close", None)):
                        yield ensure_async(close, threaded=True)
                elif callable(close := getattr(file, "close", None)):
                    close()
    return run_gen_step_iter(gen_step, async_)


@overload
def export_dir(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str], 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> int:
    ...
@overload
def export_dir(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str], 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, int]:
    ...
def export_dir(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str], 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> int | Coroutine[Any, Any, int]:
    """导出目录树

    :param client: 115 客户端或 cookies
    :param export_file_ids: 待导出的目录 id 或 pickcode
    :param target_pid: 导出到的目标目录 id 或 pickcode
    :param layer_limit: 层级深度，小于等于 0 时不限
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 返回任务 id，可用 `P115Client.fs_export_dir_status` 查询进度
    """
    if not isinstance(client, P115Client):
        client = P115Client(client, check_for_relogin=True)
    def gen_step():
        nonlocal export_file_ids, target_pid
        if isinstance(export_file_ids, int):
            pass
        elif isinstance(export_file_ids, str):
            if "," not in export_file_ids:
                export_file_ids = to_id(export_file_ids)
        else:
            export_file_ids = ",".join(str(to_id(eid)) for eid in export_file_ids)
        target_pid = to_id(target_pid)
        payload = {"file_ids": export_file_ids, "target": f"U_0_{target_pid}"}
        if layer_limit > 0:
            payload["layer_limit"] = layer_limit
        resp = yield client.fs_export_dir(payload, async_=async_, **request_kwargs)
        return check_response(resp)["data"]["export_id"]
    return run_gen_step(gen_step, async_)


@overload
def export_dir_result(
    client: str | PathLike | P115Client, 
    export_id: int | str, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def export_dir_result(
    client: str | PathLike | P115Client, 
    export_id: int | str, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def export_dir_result(
    client: str | PathLike | P115Client, 
    export_id: int | str, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """获取导出目录树的结果

    .. attention::
        如果指定超时时间为正数，则会在过期时抛出 TimeoutError，但这并不会取消远程正在执行的任务，而 115 同时只允许运行一个导出目录树的任务，所以如果要开始下一个导出任务，还需要此任务完成或者被 115 自动超时取消

    :param client: 115 客户端或 cookies
    :param export_id: 任务 id，由 `P115Client.fs_export_dir` 接口调用产生
    :param timeout: 超时秒数，如果为 None 或 小于等于 0，则相当于 float("inf")，即永不超时
    :param check_interval: 两次轮询之间的等待秒数，如果 <= 0，则不等待
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口返回结果，格式为

        .. code:: python

            {
                "export_id": str, # 任务 id
                "file_id":   str, # 导出文件的 id
                "file_name": str, # 导出文件的名字
                "pick_code": str  # 导出文件的提取码
            }
    """
    if not isinstance(client, P115Client):
        client = P115Client(client, check_for_relogin=True)
    if check_interval < 0:
        check_interval = 0
    def gen_step():
        nonlocal timeout
        if timeout is None or timeout <= 0:
            timeout = float("inf")
        do_sleep: Callable = async_sleep if async_ else sleep # type: ignore
        expired_t = time() + timeout
        while True:
            resp = yield client.fs_export_dir_status(
                export_id, 
                async_=async_, 
                **request_kwargs, 
            )
            if data := check_response(resp)["data"]:
                return data
            remaining_seconds = expired_t - time()
            if remaining_seconds <= 0:
                raise TimeoutError(export_id)
            if check_interval:
                yield do_sleep(min(check_interval, remaining_seconds))
    return run_gen_step(gen_step, async_)


@overload
def export_dir_parse_iter(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str] = 0, 
    export_id: int | str = 0, 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    parse_iter: None | Callable[[IO[bytes]], Iterator] = None, 
    delete: bool = True, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    show_clock: bool | Callable[[], Any] = False, 
    clock_interval: int | float = 0.05, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator:
    ...
@overload
def export_dir_parse_iter(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str] = 0, 
    export_id: int | str = 0, 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    parse_iter: None | Callable[[IO[bytes]], AsyncIterator] = None, 
    delete: bool = True, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    show_clock: bool | Callable[[], Any] = False, 
    clock_interval: int | float = 0.05, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator:
    ...
def export_dir_parse_iter(
    client: str | PathLike | P115Client, 
    export_file_ids: int | str | Iterable[int | str] = 0, 
    export_id: int | str = 0, 
    target_pid: int | str = 0, 
    layer_limit: int = 0, 
    parse_iter: None | Callable[[IO[bytes]], Iterator] | Callable[[IO[bytes]], AsyncIterator] = None, 
    delete: bool = True, 
    timeout: None | int | float = None, 
    check_interval: int | float = 1, 
    show_clock: bool | Callable[[], Any] = False, 
    clock_interval: int | float = 0.05, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator | AsyncIterator:
    """导出目录树到文件，读取文件并解析后返回迭代器，关闭后自动删除导出的文件

    :param client: 115 客户端或 cookies
    :param export_file_ids: 待导出的目录 id、pickcode 或 路径（如果有多个，需传入可迭代对象）
    :param export_id: 优先级高于 `export_file_ids`，之前提交的 `export_dir` 任务的 id，如果是 str，则视为导出的目录树文件的提取码（因此无需导出）
    :param target_pid: 导出到的目标目录 id 或 pickcode
    :param layer_limit: 层级深度，小于等于 0 时不限
    :param parse_iter: 解析打开的二进制文件，返回可迭代对象
    :param delete: 最终删除目录树文件（如果 export_id 为 str（即提取码），则这个值不生效，必不删除）
    :param timeout: 导出任务的超时秒数，如果为 None 或 小于等于 0，则相当于 float("inf")，即永不超时
    :param check_interval: 导出任务的状态，两次轮询之间的等待秒数，如果 <= 0，则不等待
    :param show_clock: 是否在等待导出目录树时，显示时钟。如果为 True，则显示默认的时钟，如果为 Callable，则作为自定义时钟进行调用（无参数）
    :param clock_interval: 更新时钟的时间间隔
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 解析导出文件的迭代器
    """
    if not isinstance(client, P115Client):
        client = P115Client(client, check_for_relogin=True)
    if parse_iter is None:
        if async_:
            parse_iter = partial(parse_export_dir_as_path_iter, async_=True)
        else:
            parse_iter = parse_export_dir_as_path_iter
    get_url = client.download_url
    def gen_step():
        nonlocal export_id, delete
        if not export_id:
            export_id = yield export_dir(
                client, 
                export_file_ids=export_file_ids, 
                target_pid=target_pid, 
                layer_limit=layer_limit, 
                async_=async_, 
                **request_kwargs, 
            )
        if isinstance(export_id, int):
            if not show_clock:
                result: dict = yield export_dir_result(
                    client, 
                    export_id, 
                    timeout=timeout, 
                    check_interval=check_interval, 
                    async_=async_, 
                    **request_kwargs, 
                )
            else:
                result = yield context(
                    lambda *_: export_dir_result(
                        client, 
                        export_id, 
                        timeout=timeout, 
                        check_interval=check_interval, 
                        async_=async_, 
                        **request_kwargs, 
                    ), 
                    backgroud_loop(
                        None if show_clock is True else show_clock, 
                        interval=clock_interval, 
                        async_=async_, # type: ignore
                    ), 
                    async_=async_, # type: ignore
                )
            pickcode = result["pick_code"]
        else:
            pickcode = export_id
            delete = False
        try:
            try:
                url: str = yield get_url(
                    pickcode, 
                    use_web_api=True, 
                    async_=async_, 
                    **request_kwargs, 
                )
            except OSError:
                url = yield get_url(
                    pickcode, 
                    async_=async_, 
                    **request_kwargs, 
                )
            file = client.open(url, async_=async_) # type: ignore
            try:
                if async_:
                    file_wrapper: IO = AsyncTextIOWrapper(AsyncBufferedReader(file), encoding="utf-16", newline="\n")
                else:
                    file_wrapper = TextIOWrapper(BufferedReader(file), encoding="utf-16", newline="\n")
                yield YieldFrom(parse_iter(file_wrapper)) # type: ignore
            finally:
                if async_:
                    if callable(aclose := getattr(file, "aclose", None)):
                        yield aclose()
                    elif callable(close := getattr(file, "close", None)):
                        yield ensure_async(close, threaded=True)
                elif callable(close := getattr(file, "close", None)):
                    close()
        finally:
            if delete:
                yield client.fs_delete(
                    cast(str, result["file_id"]), 
                    async_=async_, # type: ignore
                    **request_kwargs, 
                )
    return run_gen_step_iter(gen_step, async_)




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["offline_iter", "offline_restart_iter"]
__doc__ = "这个模块提供了一些和离线下载有关的函数"

from asyncio import sleep as async_sleep
from collections.abc import AsyncIterator, Callable, Iterable, Iterator
from itertools import count
from os import PathLike
from time import sleep, time
from typing import overload, Literal

from errno2 import errno
from iterutils import run_gen_step_iter, with_iter_next, Yield, YieldFrom

from ..client import check_response, P115Client, P115OpenClient
from ..exception import throw


@overload
def offline_iter(
    client: str | PathLike | P115Client | P115OpenClient, 
    /, 
    page_start: int = 1, 
    page_stop: int = -1, 
    cooldown: float = 0, 
    raise_for_update: bool = False, 
    use_open_api: bool = False,  
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def offline_iter(
    client: str | PathLike | P115Client | P115OpenClient, 
    /, 
    page_start: int = 1, 
    page_stop: int = -1, 
    cooldown: float = 0, 
    raise_for_update: bool = False, 
    use_open_api: bool = False, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def offline_iter(
    client: str | PathLike | P115Client | P115OpenClient, 
    /, 
    page_start: int = 1, 
    page_stop: int = -1, 
    cooldown: float = 0, 
    raise_for_update: bool = False, 
    use_open_api: bool = False, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[dict] | AsyncIterator[dict]:
    """遍历任务列表，获取任务信息

    .. tip::
        在逐页拉取的间隔期间，任务列表可能发生变化，可能导致重复和遗漏：

        1. 新增任务，特别是状态为进行中
        2. 删除任务
        3. 曾经取得的进行中的任务，变为完成

    :param client: 115 客户端或 cookies
    :param page_start: 开始页数
    :param page_stop: 结束页数（不含），如果 <= 0，则不限
    :param cooldown: 接口调用冷却时间，单位：秒
    :param raise_for_update: 当列表发生更新时，是否报错退出
    :param use_open_api: 是否使用 open api
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，返回任务信息
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if page_start < 1:
        page_start = 1
    if page_stop > 0:
        pages: Iterable[int] = range(page_start, page_stop)
    else:
        pages = count(page_start)
    if not isinstance(client, P115Client):
        use_open_api = True
    def gen_step():
        if use_open_api:
            offline_list = client.offline_list_open
        else:
            offline_list = client.offline_list
        may_sleep = cooldown > 0
        if may_sleep:
            do_sleep = async_sleep if async_ else sleep
            last_t: float = 0
        if raise_for_update:
            count = -1
            seen: set[str] = set()
            add_info_hash = seen.add
        for page in pages:
            if may_sleep:
                if last_t and (diff := last_t + cooldown - time()) > 0:
                    yield do_sleep(diff)
                last_t = time()
            resp = yield offline_list(page, async_=async_, **request_kwargs)
            check_response(resp)
            if use_open_api:
                resp = resp["data"]
            if raise_for_update:
                if count < 0:
                    count = resp["count"]
                elif count != resp["count"]:
                    throw(errno.EBUSY, f"detected count changes: {count} != {resp['count']}")
            tasks = resp["tasks"]
            if not tasks:
                break
            if raise_for_update:
                for task in tasks:
                    info_hash = task["info_hash"]
                    if info_hash in seen:
                        throw(errno.EBUSY, f"detected duplicate task: info_hash={info_hash!r}")
                    add_info_hash(info_hash)
                    yield Yield(task)
            else:
                yield YieldFrom(resp["tasks"])
            if len(tasks) < 30 or page >= resp["page_count"]:
                break
    return run_gen_step_iter(gen_step, async_)


@overload
def offline_restart_iter(
    client: str | PathLike | P115Client, 
    /, 
    predicate: None | Callable[[dict], bool] = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def offline_restart_iter(
    client: str | PathLike | P115Client, 
    /, 
    predicate: None | Callable[[dict], bool] = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def offline_restart_iter(
    client: str | PathLike | P115Client, 
    /, 
    predicate: None | Callable[[dict], bool] = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[dict] | AsyncIterator[dict]:
    """重试任务：重试那些因为空间不足而转存失败的任务

    :param client: 115 客户端或 cookies
    :param predicate: 断言，用于筛选
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，逐个任务返回执行重试后的响应
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    def gen_step():
        left_no_space: list[dict] = []
        add_task = left_no_space.append
        with with_iter_next(offline_iter(
            client, 
            async_=async_, 
            **request_kwargs, 
        )) as do_next:
            while True:
                task = yield do_next()
                if task["move"] == -1:
                    add_task(task)
                elif task["status"] == 2:
                    break
        for task in filter(predicate, left_no_space):
            resp = yield client.offline_restart(
                task["info_hash"], 
                async_=async_, 
                **request_kwargs, 
            )
            resp["task"] = task
            yield Yield(resp)
    return run_gen_step_iter(gen_step, async_)




#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "generate_auth_factory", "generate_cookies_factory", "generate_client_factory", 
    "make_pool", "auth_pool", "cookies_pool", "client_pool", "call_wrap_with_pool", 
]
__doc__ = "这个模块提供了一些和 cookies 池有关的函数"

from asyncio import Lock as AsyncLock
from collections.abc import Callable, Iterable, Mapping
from functools import partial, total_ordering, update_wrapper
from heapq import heappop, heappush, heapify
from itertools import cycle, repeat
from math import inf, isinf
from os import PathLike
from threading import Lock
from time import time

from http_response import get_status_code, is_timeouterror
from iterutils import run_gen_step

from ..client import check_response, P115Client
from ..exception import P115OSError, P115AuthenticationError, P115LoginError


@total_ordering
class ComparedWithID[T]:
    value: T

    def __new__(cls, value: T | ComparedWithID[T], /):
        if isinstance(value, ComparedWithID):
            return value
        else:
            self = super().__new__(cls)
            self.value = value
            return self

    def __eq__(self, other, /) -> bool:
        if isinstance(other, ComparedWithID):
            return id(self) == id(other.value)
        return id(self) == id(other)

    def __lt__(self, other, /) -> bool:
        if isinstance(other, ComparedWithID):
            return id(self) < id(other.value)
        return id(self) < id(other)

    def __repr__(self, /) -> str:
        return f"{type(self).__qualname__}({self.value!r})"


def generate_auth_factory(
    client: str | PathLike | P115Client, 
    app_ids: Iterable[int], 
    **request_kwargs, 
) -> Callable:
    """利用一个已登录设备的 cookies，产生若干开放应用的 access_token

    :param client: 115 客户端或 cookies
    :param app_ids: 一组开放应用的 AppID
    :param request_kwargs: 其它请求参数

    :return: 函数，调用以返回一个字典，包含 authorization 请求头
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    login = client.login_with_open
    get_app_id = cycle(app_ids).__next__
    def make_cookies(async_: bool = False):
        def gen_step():
            while True:
                app_id = get_app_id()
                try:
                    resp = yield login(
                        get_app_id(), 
                        async_=async_, # type: ignore
                        **request_kwargs, 
                    )
                except Exception as e:
                    if not is_timeouterror(e):
                        raise
                check_response(resp)
                return {
                    "authorization": "Bearer " + resp["data"]["access_token"], 
                    "app_id": str(app_id), 
                }
        return run_gen_step(gen_step, async_)
    return make_cookies


def generate_cookies_factory(
    client: str | PathLike | P115Client, 
    app: str | Iterable[str] = "", 
    **request_kwargs, 
) -> Callable:
    """利用一个已登录设备的 cookies，产生另一个设备的若干 cookies

    :param client: 115 客户端或 cookies
    :param app: 自动扫码后绑定的 app（多个则传入一组 app 的可迭代对象）
    :param request_kwargs: 其它请求参数

    :return: 函数，调用以返回一个字典，包含 cookie 请求头
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if isinstance(app, str):
        if app:
            if app == client.login_app():
                raise ValueError(f"same login device (app={app!r}) will cause conflicts")
        else:
            app = "tv" if client.login_ssoent == "R2" else "alipaymini"
        get_app = repeat(app).__next__
    else:
        app = tuple(app)
        if client.login_app() in app:
            raise ValueError(f"same login device (app={app!r}) will cause conflicts")
        elif not app:
            app = "tv" if client.login_ssoent == "R2" else "alipaymini"
            get_app = repeat(app).__next__
        else:
            get_app = cycle(app).__next__
    login = client.login_with_app
    def make_cookies(async_: bool = False):
        def gen_step():
            while True:
                app = get_app()
                try:
                    resp = yield login(
                        app, 
                        async_=async_, # type: ignore
                        **request_kwargs, 
                    )
                except Exception as e:
                    if not is_timeouterror(e):
                        raise
                check_response(resp)
                return {
                    "cookie": "; ".join(f"{k}={v}" for k, v in resp["data"]["cookie"].items()), 
                    "app": app, 
                }
        return run_gen_step(gen_step, async_)
    return make_cookies


def generate_client_factory(
    client: str | PathLike | P115Client, 
    app: str | Iterable[str] = "", 
    **request_kwargs, 
) -> Callable:
    """利用一个已登录设备的 client，产生另一个设备的若干 client

    :param client: 115 客户端或 cookies
    :param app: 自动扫码后绑定的 app（多个则传入一组 app 的可迭代对象）
    :param request_kwargs: 其它请求参数

    :return: 函数，调用以返回一个 client
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    cls = type(client)
    call = generate_cookies_factory(client, app, **request_kwargs)
    def make_client(async_: bool = False):
        def gen_step():
            headers = yield call(async_=async_)
            return cls(headers["cookie"])
        return run_gen_step(gen_step, async_)
    return make_client


def make_pool[T](
    generate_factory: Callable, 
    heap: None | list[tuple[float, T | ComparedWithID[T]]] = None, 
    cooldown_time: int | float = 1, 
    live_time: int | float = inf, 
    lock: bool = True, 
    **request_kwargs, 
) -> Callable:
    """创建池

    :param generate_factory: 产生值的工厂函数
    :param heap: 最小堆，可以包含一组初始值，各是一个元组，包含（上一次获取时刻, 值）
    :param cooldown_time: 值的冷却时间
    :param live_time: 值的存活时间，默认是无穷大
    :param lock: 是否需要锁
    :param request_kwargs: 其它请求参数

    :return: 返回一个函数，调用后返回一个元组，包含 值 和 一个调用以在完成后把 值 返还池中
    """
    generate = generate_factory(**request_kwargs)
    if heap is None:
        heap_: list[tuple[float,  ComparedWithID[T]]] = []
    else:
        for i, (a, b) in enumerate(heap):
            heap[i] = (a, ComparedWithID(b))
        heapify(heap)
        heap_ = heap # type: ignore
    def get_value(async_: bool = False):
        def call():
            now = time()
            if not isinf(live_time):
                watermark = now - live_time
                while heap_:
                    if heap_[0][0] > watermark:
                        break
                    heappop(heap_)
            if heap_ and heap_[0][0] + cooldown_time <= now:
                _, val = heappop(heap_)
                value = val.value
            else:
                if async_:
                    value = yield generate(async_=True)
                else:
                    value = generate()
                val = ComparedWithID(value)
            return value, partial(heappush, heap_, (time(), val))
        return run_gen_step(call, async_)
    if not lock:
        setattr(get_value, "heap", heap_)
        return get_value
    lock_sync = Lock()
    lock_async = AsyncLock()
    def locked_get_value(async_: bool = False):
        if async_:
            async def async_locked_get_value():
                async with lock_async:
                    return await get_value(async_=True)
            return async_locked_get_value
        else:
            def locked_get_value():
                with lock_sync:
                    return get_value()
            return locked_get_value
    setattr(locked_get_value, "heap", heap_)
    return locked_get_value


def auth_pool(
    client: str | PathLike | P115Client, 
    app_ids: Iterable[int], 
    heap: None | list[tuple[float, dict | ComparedWithID[dict]]] = None, 
    cooldown_time: int | float = 1, 
    live_time: int | float = 7000, 
    lock: bool = False, 
    **request_kwargs, 
) -> Callable:
    """authorization 请求头池

    :param client: 115 客户端或 cookies
    :param app_ids: 一组开放应用的 AppID
    :param heap: 最小堆，可以包含一组初始值，各是一个元组，包含（上一次获取时刻, 值）
    :param cooldown_time: 值的冷却时间
    :param live_time: 值的存活时间，默认是无穷大
    :param lock: 锁，如果不需要锁，传入 False
    :param request_kwargs: 其它请求参数

    :return: 返回一个函数，调用后返回一个元组，包含值 和 一个调用（以在完成后把值返还池中）
    """
    return make_pool(
        generate_auth_factory, 
        client=client, 
        app_ids=app_ids, 
        heap=heap, 
        cooldown_time=cooldown_time, 
        live_time=live_time, 
        lock=lock, 
        **request_kwargs, 
    )


def cookies_pool(
    client: str | PathLike | P115Client, 
    app: str | Iterable[str] = "", 
    heap: None | list[tuple[float, dict | ComparedWithID[dict]]] = None, 
    cooldown_time: int | float = 1, 
    live_time: int | float = inf, 
    lock: bool = False, 
    **request_kwargs, 
) -> Callable:
    """cookie 请求头池

    :param client: 115 客户端或 cookies
    :param app: 自动扫码后绑定的 app（多个则传入一组 app 的可迭代对象）
    :param heap: 最小堆，可以包含一组初始值，各是一个元组，包含（上一次获取时刻, 值）
    :param cooldown_time: 值的冷却时间
    :param live_time: 值的存活时间，默认是无穷大
    :param lock: 锁，如果不需要锁，传入 False
    :param request_kwargs: 其它请求参数

    :return: 返回一个函数，调用后返回一个元组，包含值 和 一个调用（以在完成后把值返还池中）
    """
    return make_pool(
        generate_cookies_factory, 
        client=client, 
        app=app, 
        heap=heap, 
        cooldown_time=cooldown_time, 
        live_time=live_time, 
        lock=lock, 
        **request_kwargs, 
    )


def client_pool(
    client: str | PathLike | P115Client, 
    app: str | Iterable[str] = "", 
    heap: None | list[tuple[float, P115Client | ComparedWithID[P115Client]]] = None, 
    cooldown_time: int | float = 1, 
    live_time: int | float = inf, 
    lock: bool = False, 
    **request_kwargs, 
) -> Callable:
    """client 池

    :param client: 115 客户端或 cookies
    :param app: 自动扫码后绑定的 app（多个则传入一组 app 的可迭代对象）
    :param heap: 最小堆，可以包含一组初始值，各是一个元组，包含（上一次获取时刻, 值）
    :param cooldown_time: 值的冷却时间
    :param live_time: 值的存活时间，默认是无穷大
    :param lock: 锁，如果不需要锁，传入 False
    :param request_kwargs: 其它请求参数

    :return: 返回一个函数，调用后返回一个元组，包含值 和 一个调用（以在完成后把值返还池中）
    """
    return make_pool(
        generate_client_factory, 
        client=client, 
        app=app, 
        heap=heap, 
        cooldown_time=cooldown_time, 
        live_time=live_time, 
        lock=lock, 
        **request_kwargs, 
    )


def call_wrap_with_pool(get_cert_headers: Callable, /, func: Callable) -> Callable:
    """包装函数，用认证信息请求头的分发池执行请求

    :param get_cert_headers: 获取认证信息的请求头的函数
    :param func: 执行请求的函数
    """
    def wrapper(
        *args, 
        headers: None | Mapping = None, 
        async_: bool = False, 
        **kwds, 
    ):
        def gen_step():
            nonlocal headers
            while True:
                if async_:
                    cert, revert = yield get_cert_headers(async_=True)
                else:
                    cert, revert = get_cert_headers()
                try:
                    if isinstance(cert, Mapping):
                        if headers:
                            headers = dict(headers, **cert)
                        else:
                            headers = cert
                        if async_:
                            resp = yield func(*args, headers=headers, async_=True, **kwds)
                        else:
                            resp = func(*args, headers=headers, **kwds)
                    elif async_:
                        resp = yield func(cert, *args, headers=headers, async_=True, **kwds)
                    else:
                        resp = func(cert, *args, headers=headers, **kwds)
                    if not isinstance(resp, dict) or resp.get("errno") != 40101004:
                        revert()
                    return resp
                except BaseException as e:
                    if isinstance(e, P115OSError) and e.args[1].get("errno") == 40101004:
                        raise
                    if not isinstance(e, (P115AuthenticationError, P115LoginError)) and get_status_code(e) != 405:
                        revert()
                        raise
        return run_gen_step(gen_step, async_)
    return update_wrapper(wrapper, func)


# TODO: 需要完整的类型签名
# TODO: 池子可以被导出，下次继续使用
# TODO: 支持多个不同设备的 cookies 组成池，以及刷新（自己刷新自己，或者由另一个 cookies 辅助刷新）
